# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

try:
    import yaml
except ImportError:
    yaml = None

import math
import unittest

import numpy as np

from lsst.sphgeom import CONTAINS, WITHIN, Angle, Circle, Ellipse, Region, UnitVector3d


class EllipseTestCase(unittest.TestCase):
    """Test Ellipse."""

    def test_construction(self):
        self.assertTrue(Ellipse.empty().isEmpty())
        self.assertTrue(Ellipse().isEmpty())
        self.assertTrue(Ellipse.full().isFull())
        e = Ellipse(Circle(UnitVector3d.X(), Angle(math.pi / 2)))
        f = Ellipse(UnitVector3d.X(), Angle(math.pi / 2))
        self.assertEqual(e, f)
        self.assertEqual(e.getAlpha(), e.getBeta())
        self.assertTrue(e.isCircle())
        self.assertTrue(e.isGreatCircle())
        g = Ellipse(e)
        h = e.clone()
        self.assertEqual(e, g)
        self.assertEqual(g, h)
        self.assertNotEqual(id(e), id(g))
        self.assertNotEqual(id(g), id(h))

    def test_comparison_operators(self):
        e = Ellipse(UnitVector3d.X(), UnitVector3d.Y(), Angle(2 * math.pi / 3))
        f = Ellipse(UnitVector3d.X(), Angle(math.pi / 3), Angle(math.pi / 6), Angle(0))
        self.assertEqual(e, e)
        self.assertNotEqual(e, f)

    def test_center_and_dimensions(self):
        e = Ellipse(UnitVector3d.X(), UnitVector3d.Y(), Angle(2 * math.pi / 3))
        self.assertAlmostEqual(e.getF1().dot(UnitVector3d.X()), 1.0)
        self.assertAlmostEqual(e.getF2().dot(UnitVector3d.Y()), 1.0)
        self.assertAlmostEqual(e.getAlpha(), Angle(2 * math.pi / 3))
        f = Ellipse(UnitVector3d.X(), Angle(math.pi / 3), Angle(math.pi / 6), Angle(0))
        self.assertEqual(f.getCenter(), UnitVector3d.X())

    def test_relationships(self):
        e = Ellipse(UnitVector3d.X(), Angle(math.pi / 3), Angle(math.pi / 6), Angle(0))
        self.assertTrue(e.contains(UnitVector3d.X()))
        self.assertTrue(UnitVector3d.X() in e)
        c = Circle(UnitVector3d.X(), Angle(math.pi / 2))
        self.assertEqual(c.relate(e), CONTAINS)
        self.assertEqual(c.overlaps(e), True)
        self.assertEqual(e.relate(c), WITHIN)
        self.assertEqual(e.overlaps(c), True)

    def test_vectorized_contains(self):
        e = Ellipse(UnitVector3d.X(), Angle(math.pi / 3), Angle(math.pi / 6), Angle(0))
        x = np.random.rand(5, 3)
        y = np.random.rand(5, 3)
        z = np.random.rand(5, 3)
        c = e.contains(x, y, z)
        lon = np.arctan2(y, x)
        lat = np.arctan2(z, np.hypot(x, y))
        c2 = e.contains(lon, lat)
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c[i, j], e.contains(u))
                self.assertEqual(c2[i, j], e.contains(u))
        # test with non-contiguous memory
        c3 = e.contains(x[::2], y[::2], z[::2])
        c4 = e.contains(lon[::2], lat[::2])
        for i in range(x.shape[0], 2):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c3[i // 2, j], e.contains(u))
                self.assertEqual(c4[i // 2, j], e.contains(u))

    def test_complement(self):
        e = Ellipse(UnitVector3d.X(), Angle(math.pi / 3), Angle(math.pi / 6), Angle(0))
        f = e.complemented().complement()
        self.assertEqual(e, f)

    def test_codec(self):
        e = Ellipse(UnitVector3d.X(), UnitVector3d.Y(), Angle(2 * math.pi / 3))
        s = e.encode()
        self.assertEqual(Ellipse.decode(s), e)
        self.assertEqual(Region.decode(s), e)

    def test_string(self):
        c = Ellipse(UnitVector3d.Z(), Angle(1.0))
        self.assertEqual(str(c), "Ellipse([0.0, 0.0, 1.0], [0.0, 0.0, 1.0], 1.0)")
        self.assertEqual(
            repr(c), "Ellipse(UnitVector3d(0.0, 0.0, 1.0), UnitVector3d(0.0, 0.0, 1.0), Angle(1.0))"
        )
        self.assertEqual(c, eval(repr(c), {"Angle": Angle, "Ellipse": Ellipse, "UnitVector3d": UnitVector3d}))

    def test_pickle(self):
        a = Ellipse(UnitVector3d.X(), UnitVector3d.Y(), Angle(2 * math.pi / 3))
        b = pickle.loads(pickle.dumps(a, pickle.HIGHEST_PROTOCOL))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def test_yaml(self):
        a = Ellipse(UnitVector3d.X(), UnitVector3d.Y(), Angle(2 * math.pi / 3))
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

try:
    import yaml
except ImportError:
    yaml = None

import unittest

import numpy as np

from lsst.sphgeom import CONTAINS, Circle, ConvexPolygon, Region, UnitVector3d


class ConvexPolygonTestCase(unittest.TestCase):
    """Test ConvexPolygon."""

    def testConstruction(self):
        points = [UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()]
        p1 = ConvexPolygon(points)
        self.assertEqual(points, p1.getVertices())
        p2 = p1.clone()
        self.assertEqual(p1, p2)
        p3 = ConvexPolygon([-UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        self.assertNotEqual(p1, p3)
        p4 = ConvexPolygon.convexHull(
            [UnitVector3d.Y(), UnitVector3d.X(), UnitVector3d(1, 1, 1), UnitVector3d.Z()]
        )
        self.assertEqual(p1, p4)

    def testCodec(self):
        p = ConvexPolygon([UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        s = p.encode()
        self.assertEqual(ConvexPolygon.decode(s), p)
        self.assertEqual(Region.decode(s), p)

    def testRelationships(self):
        p = ConvexPolygon([UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        self.assertTrue(p.contains(p.getCentroid()))
        boundingCircle = p.getBoundingCircle()
        self.assertEqual(boundingCircle.relate(p), CONTAINS)
        self.assertTrue(p.isWithin(boundingCircle))
        self.assertTrue(p.intersects(boundingCircle))
        self.assertFalse(p.isDisjointFrom(boundingCircle))
        self.assertEqual(p.overlaps(boundingCircle), True)
        self.assertFalse(p.contains(boundingCircle))
        tinyCircle = Circle(boundingCircle.getCenter())
        self.assertFalse(p.isWithin(tinyCircle))
        self.assertTrue(p.intersects(tinyCircle))
        self.assertFalse(p.isDisjointFrom(tinyCircle))
        self.assertEqual(p.overlaps(tinyCircle), True)
        self.assertTrue(p.contains(tinyCircle))

    def test_vectorized_contains(self):
        b = ConvexPolygon([UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        x = np.random.rand(5, 3)
        y = np.random.rand(5, 3)
        z = np.random.rand(5, 3)
        c = b.contains(x, y, z)
        lon = np.arctan2(y, x)
        lat = np.arctan2(z, np.hypot(x, y))
        c2 = b.contains(lon, lat)
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c[i, j], b.contains(u))
                self.assertEqual(c2[i, j], b.contains(u))
        # test with non-contiguous memory
        c3 = b.contains(x[::2], y[::2], z[::2])
        c4 = b.contains(lon[::2], lat[::2])
        for i in range(x.shape[0], 2):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c3[i // 2, j], b.contains(u))
                self.assertEqual(c4[i // 2, j], b.contains(u))

    def testString(self):
        p = ConvexPolygon([UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        self.assertEqual(str(p), repr(p))
        self.assertEqual(
            repr(p),
            "ConvexPolygon([UnitVector3d(0.0, 0.0, 1.0), "
            "UnitVector3d(1.0, 0.0, 0.0), "
            "UnitVector3d(0.0, 1.0, 0.0)])",
        )
        self.assertEqual(p, eval(repr(p), {"ConvexPolygon": ConvexPolygon, "UnitVector3d": UnitVector3d}))

    def testPickle(self):
        a = ConvexPolygon([UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        b = pickle.loads(pickle.dumps(a, pickle.HIGHEST_PROTOCOL))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def testYaml(self):
        a = ConvexPolygon([UnitVector3d.Z(), UnitVector3d.X(), UnitVector3d.Y()])
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from lsst.sphgeom import Box, Circle, ConvexPolygon, LonLat, Region, UnitVector3d


class IvoaTestCase(unittest.TestCase):
    """Test Box."""

    def test_construction(self):
        # Example POS strings found in IVOA documentation.
        example_pos = (
            "CIRCLE 12.0 34.0 0.5",
            "RANGE 12.0 12.5 34.0 36.0",
            "POLYGON 12.0 34.0 14.0 35.0 14. 36.0 12.0 35.0",
            "RANGE 0 360.0 -2.0 2.0",
            "RANGE 0 360.0 89.0 +Inf",
            "RANGE -Inf +Inf -Inf +Inf",
            "POLYGON 12 34 14 34 14 36 12 36",
            "RANGE 0 360 89 90",
        )
        for pos in example_pos:
            region = Region.from_ivoa_pos(pos)
            self.assertIsInstance(region, Region)

        # Badly formed strings raising ValueError.
        bad_pos = (
            "circle 12 34 0.5",
            "CIRCLE 12 34 1 1",
            "RANGE 0 360",
            "POLYGON 0 1 2 3",
            "POLYGON 0 1 2 3 4 5 6",
            "CONVEXPOLYGON 0 1 2 3 4 5",
        )
        for pos in bad_pos:
            with self.assertRaises(ValueError):
                Region.from_ivoa_pos(pos)

    def _split_pos(self, pos: str):
        """Split POS into type and floats."""
        region_type, *coordstr = pos.split()
        coordinates = [float(c) for c in coordstr]
        return region_type, coordinates

    def assert_pos_equal(self, pos1: str, pos2: str):
        """Compare two POS strings and check for equality taking into account
        floating point differences.
        """
        region_type1, coords1 = self._split_pos(pos1)
        region_type2, coords2 = self._split_pos(pos2)
        self.assertEqual(region_type1, region_type2)
        self.assertEqual(len(coords1), len(coords2))
        for c1, c2 in zip(coords1, coords2, strict=True):
            self.assertAlmostEqual(c1, c2)

    def test_circle(self):
        """Test circle construction."""
        pos = "CIRCLE 12.0 34.0 5"
        circle = Region.from_ivoa_pos(pos)
        self.assertIsInstance(circle, Circle)
        self.assertTrue(circle.contains(UnitVector3d(LonLat.fromDegrees(13.0, 33.0))))
        self.assertFalse(circle.contains(UnitVector3d(LonLat.fromDegrees(12.0, 40.0))))

        self.assert_pos_equal(circle.to_ivoa_pos(), pos)

    def test_range(self):
        """Test range construction."""
        box = Region.from_ivoa_pos("RANGE 1 2 5 6")
        self.assertIsInstance(box, Box)
        self.assertTrue(box.contains(UnitVector3d(LonLat.fromDegrees(1.5, 5.4))))
        self.assertFalse(box.contains(UnitVector3d(LonLat.fromDegrees(4, 10))))
        self.assert_pos_equal(box.to_ivoa_pos(), "RANGE 1 2 5 6")

        box = Region.from_ivoa_pos("RANGE 1 2 20 +Inf")
        self.assertTrue(box.contains(UnitVector3d(LonLat.fromDegrees(1.7, 80))))
        self.assertFalse(box.contains(UnitVector3d(LonLat.fromDegrees(1.7, 10))))
        self.assert_pos_equal(box.to_ivoa_pos(), "RANGE 1 2 20 90")

        box = Region.from_ivoa_pos("RANGE 50 +Inf 20 30")
        self.assertTrue(box.contains(UnitVector3d(LonLat.fromDegrees(60, 25))))
        self.assertFalse(box.contains(UnitVector3d(LonLat.fromDegrees(49, 21))))
        self.assert_pos_equal(box.to_ivoa_pos(), "RANGE 50 360 20 30")

        box = Region.from_ivoa_pos("RANGE -Inf +50 20 30")
        self.assertTrue(box.contains(UnitVector3d(LonLat.fromDegrees(40, 25))))
        self.assertFalse(box.contains(UnitVector3d(LonLat.fromDegrees(60, 21))))
        self.assert_pos_equal(box.to_ivoa_pos(), "RANGE 0 50 20 30")

        box = Region.from_ivoa_pos("RANGE -Inf +Inf 20 30")
        self.assertTrue(box.contains(UnitVector3d(LonLat.fromDegrees(10, 25))))
        self.assertTrue(box.contains(UnitVector3d(LonLat.fromDegrees(359, 25))))
        self.assertFalse(box.contains(UnitVector3d(LonLat.fromDegrees(49, 19))))
        self.assert_pos_equal(box.to_ivoa_pos(), "RANGE 0 360. 20 30")

    def test_polygon(self):
        """Test polygon construction."""
        pos = "POLYGON 12.0 34.0 14.0 35.0 14. 36.0 12.0 35.0"
        poly = Region.from_ivoa_pos(pos)
        self.assertIsInstance(poly, ConvexPolygon)
        self.assertTrue(poly.contains(UnitVector3d(LonLat.fromDegrees(13, 35))))
        self.assertFalse(poly.contains(UnitVector3d(LonLat.fromDegrees(14, 34))))
        self.assert_pos_equal(poly.to_ivoa_pos(), pos)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

try:
    import yaml
except ImportError:
    yaml = None

import unittest

from lsst.sphgeom import Angle, Circle, Mq3cPixelization, RangeSet, UnitVector3d


class Mq3cPixelizationTestCase(unittest.TestCase):
    """Test MQ3C pixelization."""

    def test_construction(self):
        with self.assertRaises(ValueError):
            Mq3cPixelization(-1)
        with self.assertRaises(ValueError):
            Mq3cPixelization(Mq3cPixelization.MAX_LEVEL + 1)
        m1 = Mq3cPixelization(0)
        self.assertEqual(m1.getLevel(), 0)
        m2 = Mq3cPixelization(1)
        m3 = Mq3cPixelization(m2)
        self.assertNotEqual(m1, m2)
        self.assertEqual(m2, m3)

    def test_indexing(self):
        pixelization = Mq3cPixelization(1)
        self.assertEqual(pixelization.index(UnitVector3d(0.5, -0.5, 1.0)), 53)

    def test_level(self):
        self.assertEqual(Mq3cPixelization.level(0), -1)
        for level in range(Mq3cPixelization.MAX_LEVEL + 1):
            for root in range(8, 10):
                index = root * 4**level
                self.assertEqual(Mq3cPixelization.level(index), -1)
            for root in range(10, 16):
                index = root * 4**level
                self.assertEqual(Mq3cPixelization.level(index), level)

    def test_envelope_and_interior(self):
        pixelization = Mq3cPixelization(1)
        c = Circle(UnitVector3d(1.0, -0.5, -0.5), Angle.fromDegrees(0.1))
        rs = pixelization.envelope(c)
        self.assertTrue(rs == RangeSet(44))
        rs = pixelization.envelope(c, 1)
        self.assertTrue(rs == RangeSet(44))
        self.assertTrue(rs.isWithin(pixelization.universe()))
        rs = pixelization.interior(c)
        self.assertTrue(rs.empty())

    def test_index_to_string(self):
        strings = ["+X", "+Y", "+Z", "-X", "-Y", "-Z"]
        for i in range(6):
            s0 = strings[i]
            components = [0.0] * 3
            components[i % 3] = 1.0 if i < 3 else -1.0
            v = UnitVector3d(*components)
            f = Mq3cPixelization(0).index(v)
            self.assertEqual(Mq3cPixelization.asString(f), s0)
            self.assertEqual(Mq3cPixelization(0).toString(f), s0)
            for j in range(4):
                s1 = s0 + str(j)
                self.assertEqual(Mq3cPixelization.asString(f * 4 + j), s1)
                self.assertEqual(Mq3cPixelization(1).toString(f * 4 + j), s1)

    def test_string(self):
        p = Mq3cPixelization(3)
        self.assertEqual(str(p), "Mq3cPixelization(3)")
        self.assertEqual(str(p), repr(p))
        self.assertEqual(p, eval(repr(p), {"Mq3cPixelization": Mq3cPixelization}))

    def test_pickle(self):
        a = Mq3cPixelization(20)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def test_yaml(self):
        a = Mq3cPixelization(20)
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

try:
    import yaml
except ImportError:
    yaml = None

import math
import unittest

import numpy as np

from lsst.sphgeom import (
    CONTAINS,
    DISJOINT,
    Angle,
    AngleInterval,
    Box,
    LonLat,
    NormalizedAngle,
    NormalizedAngleInterval,
    Region,
    UnitVector3d,
)


class BoxTestCase(unittest.TestCase):
    """Test Box."""

    def setUp(self):
        np.random.seed(1)

    def test_construction(self):
        b = Box(Box.allLongitudes(), Box.allLatitudes())
        self.assertTrue(b.isFull())
        b = Box.fromDegrees(-90, -45, 90, 45)
        self.assertEqual(b, Box(b.getLon(), b.getLat()))
        a = Box.fromRadians(-0.5 * math.pi, -0.25 * math.pi, 0.5 * math.pi, 0.25 * math.pi)
        b = Box(
            LonLat.fromRadians(-0.5 * math.pi, -0.25 * math.pi),
            LonLat.fromRadians(0.5 * math.pi, 0.25 * math.pi),
        )
        c = Box(LonLat.fromRadians(0, 0), Angle(0.5 * math.pi), Angle(0.25 * math.pi))
        d = c.clone()
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        self.assertEqual(c, d)
        self.assertNotEqual(id(c), id(d))
        b = Box()
        self.assertTrue(b.isEmpty())
        self.assertTrue(Box.empty().isEmpty())
        self.assertTrue(Box.full().isFull())

    def test_comparison_operators(self):
        self.assertEqual(Box(LonLat.fromDegrees(45, 45)), LonLat.fromDegrees(45, 45))
        self.assertEqual(
            Box.fromDegrees(90, -45, 180, 45),
            Box(NormalizedAngleInterval.fromDegrees(90, 180), AngleInterval.fromDegrees(-45, 45)),
        )
        self.assertNotEqual(Box(LonLat.fromDegrees(45, 45)), LonLat.fromDegrees(45, 90))
        self.assertNotEqual(Box.fromDegrees(90, -45, 180, 45), Box.fromDegrees(90, -45, 180, 90))

    def test_center_and_dimensions(self):
        b = Box.fromDegrees(-90, -45, 90, 45)
        self.assertEqual(b.getCenter(), LonLat.fromDegrees(0, 0))
        self.assertEqual(b.getWidth(), Angle.fromDegrees(180))
        self.assertEqual(b.getHeight(), Angle.fromDegrees(90))
        self.assertEqual(b.getLon().getA(), NormalizedAngle.fromDegrees(-90))
        self.assertEqual(b.getLat().getB(), Angle.fromDegrees(45))

    def test_relationships(self):
        b1 = Box.fromDegrees(90, 0, 180, 45)
        p = LonLat.fromDegrees(135, 10)
        self.assertTrue(p in b1)
        self.assertTrue(b1.contains(p))
        b2 = Box.fromDegrees(135, 15, 135, 30)
        self.assertTrue(b1.contains(b2))
        self.assertTrue(b2.isWithin(b1))
        b3 = Box.fromDegrees(0, -45, 90, 0)
        u = UnitVector3d(1, 1, -1)
        self.assertTrue(b1.intersects(b3))
        self.assertTrue(u in b3)
        self.assertTrue(b3.contains(u))
        b4 = Box.fromDegrees(200, 10, 300, 20)
        self.assertTrue(b1.isDisjointFrom(b4))
        self.assertEqual(b1.overlaps(b4), False)
        r = b1.relate(LonLat.fromDegrees(135, 10))
        self.assertEqual(r, CONTAINS)
        r = b4.relate(b1)
        self.assertEqual(r, DISJOINT)
        self.assertEqual(b4.overlaps(b1), False)

    def test_vectorized_contains(self):
        b = Box.fromDegrees(200, 10, 300, 20)
        x = np.random.rand(5, 3)
        y = np.random.rand(5, 3)
        z = np.random.rand(5, 3)
        c = b.contains(x, y, z)
        lon = np.arctan2(y, x)
        lat = np.arctan2(z, np.hypot(x, y))
        c2 = b.contains(lon, lat)
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c[i, j], b.contains(u))
                self.assertEqual(c2[i, j], b.contains(u))
        # test with non-contiguous memory
        c3 = b.contains(x[::2], y[::2], z[::2])
        c4 = b.contains(lon[::2], lat[::2])
        for i in range(x.shape[0], 2):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c3[i // 2, j], b.contains(u))
                self.assertEqual(c4[i // 2, j], b.contains(u))

    def test_expanding_and_clipping(self):
        a = Box.fromDegrees(0, 0, 10, 10)
        b = (
            a.expandedTo(LonLat.fromDegrees(20, 20))
            .expandedTo(Box.fromDegrees(0, 0, 30, 10))
            .clippedTo(Box.fromDegrees(10, 10, 15, 15))
            .clippedTo(LonLat.fromDegrees(11, 11))
        )
        a.expandTo(LonLat.fromDegrees(20, 20))
        a.expandTo(Box.fromDegrees(0, 0, 30, 10))
        a.clipTo(Box.fromDegrees(10, 10, 15, 15))
        a.clipTo(LonLat.fromDegrees(11, 11))
        self.assertEqual(a, b)
        self.assertEqual(a, LonLat.fromDegrees(11, 11))
        a.clipTo(LonLat.fromDegrees(0, 0))
        self.assertTrue(a.isEmpty())

    def test_dilation_and_erosion(self):
        a = Box.fromRadians(0.5, -0.5, 1.5, 0.5)
        b = a.dilatedBy(Angle(0.5), Angle(0.5)).erodedBy(Angle(1), Angle(1))
        a.dilateBy(Angle(0.5), Angle(0.5)).erodeBy(Angle(1), Angle(1))
        self.assertEqual(a, b)
        self.assertEqual(a, LonLat.fromRadians(1, 0))

    def test_codec(self):
        b = Box.fromRadians(0, 0, 1, 1)
        s = b.encode()
        self.assertEqual(Box.decode(s), b)
        self.assertEqual(Region.decode(s), b)

    def test_string(self):
        b = Box.fromRadians(0, 0, 1, 1)
        self.assertEqual(str(b), "Box([0.0, 1.0], [0.0, 1.0])")
        self.assertEqual(
            repr(b),
            "Box(NormalizedAngleInterval.fromRadians(0.0, 1.0), AngleInterval.fromRadians(0.0, 1.0))",
        )
        self.assertEqual(
            b,
            eval(
                repr(b),
                {
                    "AngleInterval": AngleInterval,
                    "Box": Box,
                    "NormalizedAngleInterval": NormalizedAngleInterval,
                },
            ),
        )

    def test_pickle(self):
        a = Box.fromDegrees(0, 0, 10, 10)
        b = pickle.loads(pickle.dumps(a, pickle.HIGHEST_PROTOCOL))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def test_yaml(self):
        a = Box.fromDegrees(0, 0, 10, 10)
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest
from base64 import b64encode

try:
    import yaml
except ImportError:
    yaml = None

from lsst.sphgeom import (
    CONTAINS,
    DISJOINT,
    INTERSECTS,
    WITHIN,
    Angle,
    AngleInterval,
    Box,
    Circle,
    CompoundRegion,
    IntersectionRegion,
    LonLat,
    NormalizedAngleInterval,
    Region,
    UnionRegion,
    UnitVector3d,
)


class CompoundRegionTestMixin:
    """Tests for both UnionRegion and IntersectionRegion.

    Concrete TestCase subclasses are responsible for adding an `instance`
    attribute with a non-trivial instance of the `CompoundRegion` subclass
    being tested.
    """

    def setUp(self):
        self.point_in_circle = LonLat.fromDegrees(44.0, 45.0)
        self.point_in_box = LonLat.fromDegrees(46.0, 45.0)
        self.point_in_both = LonLat.fromDegrees(45.0, 45.0)
        self.point_in_neither = LonLat.fromDegrees(45.0, 48.0)
        self.circle = Circle(UnitVector3d(self.point_in_circle), Angle.fromDegrees(1.0))
        self.box = Box.fromDegrees(
            self.point_in_box.getLon().asDegrees() - 1.5,
            self.point_in_box.getLat().asDegrees() - 1.5,
            self.point_in_box.getLon().asDegrees() + 1.5,
            self.point_in_box.getLat().asDegrees() + 1.5,
        )
        self.faraway = Circle(UnitVector3d(self.point_in_neither), Angle.fromDegrees(0.1))
        self.operands = (self.circle, self.box)

    def assertOperandsEqual(self, region, operands):
        """Assert that a compound regions operands are equal to the given
        tuple of operands.
        """
        regions = tuple(region.cloneOperand(i) for i in range(region.nOperands()))
        self.assertCountEqual(regions, operands)

    def assertCompoundRegionsEqual(self, a, b):
        """Assert that two compound regions are equal.

        CompoundRegion does not implement equality comparison because
        regions in general do not, and hence it cannot delegate that operation
        to its operands.  But the concrete operands (circle and box) we use in
        these tests do implement equality comparison.
        """
        operands = tuple(b.cloneOperand(i) for i in range(b.nOperands()))
        self.assertEqual(type(a), type(b))
        self.assertOperandsEqual(a, operands)

    def assertRelations(self, r1, r2, relation, overlaps):
        """Assert relation between two regions."""
        self.assertEqual(r1.relate(r2), relation)
        self.assertEqual(r1.overlaps(r2), overlaps)

    def testSetUp(self):
        """Test that the points and operand regions being tested have the
        relationships expected.
        """
        self.assertTrue(self.circle.contains(UnitVector3d(self.point_in_circle)))
        self.assertTrue(self.circle.contains(UnitVector3d(self.point_in_both)))
        self.assertFalse(self.circle.contains(UnitVector3d(self.point_in_box)))
        self.assertFalse(self.circle.contains(UnitVector3d(self.point_in_neither)))
        self.assertTrue(self.box.contains(UnitVector3d(self.point_in_box)))
        self.assertTrue(self.box.contains(UnitVector3d(self.point_in_both)))
        self.assertFalse(self.box.contains(UnitVector3d(self.point_in_circle)))
        self.assertFalse(self.box.contains(UnitVector3d(self.point_in_neither)))
        self.assertRelations(self.circle, self.circle, CONTAINS | WITHIN, True)
        self.assertRelations(self.circle, self.box, INTERSECTS, True)
        self.assertRelations(self.circle, self.faraway, DISJOINT, False)
        self.assertRelations(self.box, self.circle, INTERSECTS, True)
        self.assertRelations(self.box, self.box, CONTAINS | WITHIN, True)
        self.assertRelations(self.box, self.faraway, DISJOINT, False)

    def testOperands(self):
        """Test the cloneOperands accessor."""
        self.assertOperandsEqual(self.instance, self.operands)

    def testIterator(self):
        """Test Python iteration."""
        self.assertEqual(len(self.instance), len(self.operands))
        it = iter(self.instance)
        self.assertEqual(next(it), self.operands[0])
        self.assertEqual(next(it), self.operands[1])
        with self.assertRaises(StopIteration):
            next(it)

    def testCodec(self):
        """Test that encode and decode round-trip."""
        s = self.instance.encode()
        self.assertCompoundRegionsEqual(type(self.instance).decode(s), self.instance)
        self.assertCompoundRegionsEqual(CompoundRegion.decode(s), self.instance)
        self.assertCompoundRegionsEqual(Region.decode(s), self.instance)

    def testPickle(self):
        """Test pickling round-trips."""
        s = pickle.dumps(self.instance, pickle.HIGHEST_PROTOCOL)
        self.assertCompoundRegionsEqual(pickle.loads(s), self.instance)

    def testString(self):
        """Test that repr returns a string that can be eval'd to yield an
        equivalent instance.
        """
        self.assertCompoundRegionsEqual(
            self.instance,
            eval(
                repr(self.instance),
                {
                    "UnionRegion": UnionRegion,
                    "IntersectionRegion": IntersectionRegion,
                    "Box": Box,
                    "Circle": Circle,
                    "UnitVector3d": UnitVector3d,
                    "Angle": Angle,
                    "AngleInterval": AngleInterval,
                    "NormalizedAngleInterval": NormalizedAngleInterval,
                },
            ),
        )

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def testYaml(self):
        """Test that YAML dump and load round-trip."""
        self.assertCompoundRegionsEqual(yaml.safe_load(yaml.dump(self.instance)), self.instance)


class UnionRegionTestCase(CompoundRegionTestMixin, unittest.TestCase):
    """Test UnionRegion."""

    def setUp(self):
        CompoundRegionTestMixin.setUp(self)
        self.instance = UnionRegion(*self.operands)

    def testEmpty(self):
        """Test zero-operand union which is quivalent to empty region."""
        region = UnionRegion()

        self.assertFalse(region.contains(UnitVector3d(self.point_in_both)))
        self.assertFalse(region.contains(UnitVector3d(self.point_in_circle)))
        self.assertFalse(region.contains(UnitVector3d(self.point_in_box)))
        self.assertFalse(region.contains(UnitVector3d(self.point_in_neither)))

        self.assertRelations(region, self.box, DISJOINT, False)
        self.assertRelations(region, self.circle, DISJOINT, False)
        self.assertRelations(region, self.faraway, DISJOINT, False)
        self.assertRelations(region, self.instance, DISJOINT, False)
        self.assertRelations(self.box, region, DISJOINT, False)
        self.assertRelations(self.circle, region, DISJOINT, False)
        self.assertRelations(self.faraway, region, DISJOINT, False)
        self.assertRelations(self.instance, region, DISJOINT, False)

        self.assertEqual(Region.getRegions(region), [])

    def testContains(self):
        """Test point-in-region checks."""
        self.assertTrue(self.instance.contains(UnitVector3d(self.point_in_both)))
        self.assertTrue(self.instance.contains(UnitVector3d(self.point_in_circle)))
        self.assertTrue(self.instance.contains(UnitVector3d(self.point_in_box)))
        self.assertFalse(self.instance.contains(UnitVector3d(self.point_in_neither)))

    def testRelate(self):
        """Test region-region relationship checks."""
        self.assertRelations(self.instance, self.circle, CONTAINS, True)
        self.assertRelations(self.instance, self.box, CONTAINS, True)
        self.assertRelations(self.instance, self.faraway, DISJOINT, False)
        self.assertRelations(self.circle, self.instance, WITHIN, True)
        self.assertRelations(self.box, self.instance, WITHIN, True)
        self.assertRelations(self.faraway, self.instance, DISJOINT, False)

    def testBounding(self):
        """Test for getBounding*() methods."""
        region = UnionRegion()
        self.assertTrue(region.getBoundingBox().empty())
        self.assertTrue(region.getBoundingBox3d().empty())
        self.assertTrue(region.getBoundingCircle().empty())

        for operand in self.operands:
            self.assertEqual(self.instance.getBoundingBox().relate(operand), CONTAINS)
        for operand in self.operands:
            self.assertTrue(self.instance.getBoundingBox3d().contains(operand.getBoundingBox3d()))
        # This test fails for first operand (Circle), I guess due to precision.
        for operand in self.operands[-1:]:
            self.assertTrue(self.instance.getBoundingCircle().relate(operand), CONTAINS)

    def testDecodeBase64(self):
        """Test Region.decodeBase64, which includes special handling for
        union regions.
        """
        # Test with the full UnionRegion encoded, then base64-encoded.
        s1 = b64encode(self.instance.encode()).decode("ascii")
        self.assertCompoundRegionsEqual(Region.decodeBase64(s1), self.instance)
        # Test alternate form with union members concatenated with ':' after
        # base64-encoding.
        s2 = ":".join(b64encode(region.encode()).decode("ascii") for region in self.instance)
        self.assertCompoundRegionsEqual(Region.decodeBase64(s2), self.instance)
        # Test that an empty string decodes as a UnionRegion with no members.
        self.assertCompoundRegionsEqual(Region.decodeBase64(""), UnionRegion())


class IntersectionRegionTestCase(CompoundRegionTestMixin, unittest.TestCase):
    """Test intersection region."""

    def setUp(self):
        CompoundRegionTestMixin.setUp(self)
        self.instance = IntersectionRegion(*self.operands)

    def testEmpty(self):
        """Test zero-operand intersection (equivalent to full sphere)."""
        region = IntersectionRegion()

        self.assertTrue(region.contains(UnitVector3d(self.point_in_both)))
        self.assertTrue(region.contains(UnitVector3d(self.point_in_circle)))
        self.assertTrue(region.contains(UnitVector3d(self.point_in_box)))
        self.assertTrue(region.contains(UnitVector3d(self.point_in_neither)))

        self.assertRelations(region, self.box, CONTAINS, True)
        self.assertRelations(region, self.circle, CONTAINS, True)
        self.assertRelations(region, self.faraway, CONTAINS, True)
        self.assertRelations(region, self.instance, CONTAINS, True)
        self.assertRelations(self.box, region, WITHIN, True)
        self.assertRelations(self.circle, region, WITHIN, True)
        self.assertRelations(self.faraway, region, WITHIN, True)
        # Overlaps between intersections are very conservative.
        self.assertRelations(self.instance, region, WITHIN, None)

        self.assertEqual(Region.getRegions(region), [])

    def testContains(self):
        """Test point-in-region checks."""
        self.assertTrue(self.instance.contains(UnitVector3d(self.point_in_both)))
        self.assertFalse(self.instance.contains(UnitVector3d(self.point_in_circle)))
        self.assertFalse(self.instance.contains(UnitVector3d(self.point_in_box)))
        self.assertFalse(self.instance.contains(UnitVector3d(self.point_in_neither)))

    def testRelate(self):
        """Test region-region relationship checks."""
        self.assertRelations(self.instance, self.box, WITHIN, None)
        self.assertRelations(self.instance, self.circle, WITHIN, None)
        self.assertRelations(self.instance, self.faraway, DISJOINT, False)
        self.assertRelations(self.circle, self.instance, CONTAINS, None)
        self.assertRelations(self.box, self.instance, CONTAINS, None)
        self.assertRelations(self.faraway, self.instance, DISJOINT, False)

    def testGetRegion(self):
        c1 = Circle(UnitVector3d(0.0, 0.0, 1.0), 1.0)
        c2 = Circle(UnitVector3d(1.0, 0.0, 1.0), 2.0)
        b1 = Box.fromDegrees(90, 0, 180, 45)
        b2 = Box.fromDegrees(135, 15, 135, 30)
        u1 = UnionRegion(c1, b1)
        u2 = UnionRegion(c2, b2)
        i1 = IntersectionRegion(c1, b1)
        i2 = IntersectionRegion(c2, b2)
        ur = UnionRegion(u1, u2)
        ir = IntersectionRegion(i1, i2)
        self.assertEqual(Region.getRegions(c1), [c1])
        self.assertEqual(Region.getRegions(i1), [c1, b1])
        self.assertEqual(Region.getRegions(u1), [c1, b1])
        # Compounds of compounds will be flattened, order preserved.
        self.assertEqual(Region.getRegions(ir), [c1, b1, c2, b2])
        self.assertEqual(Region.getRegions(ur), [c1, b1, c2, b2])

        # TODO: This test fails because CompoundRegion does not define
        # equality operator, and it is non-trivial to add one.
        # ur2 = UnionRegion(u1, i1, u2)
        # self.assertEqual(Region.getRegions(ur2), [c1, b1, i1, c2, b2])

    def testBounding(self):
        """Test for getBounding*() methods."""
        region = UnionRegion()
        self.assertTrue(region.getBoundingBox().full())
        self.assertTrue(region.getBoundingBox3d().full())
        self.assertTrue(region.getBoundingCircle().full())

        # Only Box3d test works reliably, other two fails due to boundary
        # overlaps and precision.
        for operand in self.operands:
            self.assertTrue(operand.getBoundingBox3d().contains(self.instance.getBoundingBox3d()))

    def testDecodeOverlapsBase64(self):
        """Test Region.decodeOverlapsBase64.

        This test is in this test case because it can make good use of the
        concrete regions defined in setUp.
        """

        def run_overlaps(pairs):
            or_terms = []
            for a, b in pairs:
                a_str = b64encode(a.encode()).decode("ascii")
                b_str = b64encode(b.encode()).decode("ascii")
                or_terms.append(f"{a_str}&{b_str}")
            overlap_str = "|".join(or_terms)
            return Region.decodeOverlapsBase64(overlap_str)

        self.assertEqual(run_overlaps([]), False)
        self.assertEqual(run_overlaps([(self.box, self.circle)]), True)
        self.assertEqual(run_overlaps([(self.box, self.faraway)]), False)
        self.assertEqual(run_overlaps([(self.circle, self.faraway)]), False)
        self.assertEqual(run_overlaps([(self.instance, self.box)]), None)
        self.assertEqual(run_overlaps([(self.box, self.circle), (self.box, self.faraway)]), True)
        self.assertEqual(run_overlaps([(self.faraway, self.circle), (self.box, self.faraway)]), False)
        self.assertEqual(run_overlaps([(self.instance, self.box), (self.circle, self.faraway)]), None)
        self.assertEqual(run_overlaps([(self.instance, self.box), (self.circle, self.box)]), True)
        self.assertEqual(run_overlaps([(self.circle, self.box), (self.instance, self.box)]), True)

        with self.assertRaises(RuntimeError):
            # Decoding a single region is an error; that's not an expression.
            Region.decodeOverlapsBase64(b64encode(self.box.encode()).decode("ascii"))


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import RangeSet


class RangeSetTestCase(unittest.TestCase):
    """Test RangeSet."""

    def testConstruction(self):
        s1 = RangeSet(1)
        s2 = RangeSet()
        s3 = RangeSet(2, 1)
        s4 = RangeSet(s3)
        self.assertTrue(s2.empty())
        self.assertEqual(s3, s4)
        self.assertEqual(s1, s3.complement())

    def testComparisonOperators(self):
        s1 = RangeSet(1)
        s2 = RangeSet(2)
        self.assertNotEqual(s1, s2)
        s1.insert(2)
        s2.insert(1)
        self.assertEqual(s1, s2)
        self.assertTrue(RangeSet(2, 1).contains(RangeSet(3, 4)))
        self.assertTrue(RangeSet(2, 1).contains(3, 4))
        self.assertTrue(RangeSet(2, 1).contains(3))
        self.assertTrue(RangeSet(2, 4).isWithin(RangeSet(1, 5)))
        self.assertTrue(RangeSet(2, 4).isWithin(1, 5))
        self.assertFalse(RangeSet(2, 4).isWithin(3))
        self.assertTrue(RangeSet(2, 4).intersects(RangeSet(3, 5)))
        self.assertTrue(RangeSet(2, 4).intersects(3, 5))
        self.assertTrue(RangeSet(2, 4).intersects(3))
        self.assertTrue(RangeSet(2, 4).isDisjointFrom(RangeSet(6, 8)))
        self.assertTrue(RangeSet(2, 4).isDisjointFrom(6, 8))
        self.assertTrue(RangeSet(2, 4).isDisjointFrom(6))

    def testSetOperators(self):
        a = RangeSet(1)
        b = ~a
        self.assertTrue((a | b).full())
        self.assertTrue((a & b).empty())
        self.assertEqual(a - b, a)
        self.assertEqual(b - a, b)
        a &= a
        b &= b
        c = (a ^ b) - RangeSet(2, 4)
        self.assertEqual(c, RangeSet(4, 2))
        c |= b
        self.assertTrue(c.full())
        c ^= c
        self.assertTrue(c.empty())

    def testRanges(self):
        s = RangeSet()
        s.insert(0, 1)
        s.insert(2, 3)
        self.assertEqual(s.ranges(), [(0, 1), (2, 3)])
        s = RangeSet(4, 2)
        self.assertEqual(list(s), [(0, 2), (4, 0)])

    def testString(self):
        s = RangeSet(1, 10)
        self.assertEqual(str(s), "[(1, 10)]")
        self.assertEqual(repr(s), "RangeSet([(1, 10)])")
        self.assertEqual(s, eval(repr(s), {"RangeSet": RangeSet}))

    def testPickle(self):
        r = RangeSet([2, 3, 5, 7, 11, 13, 17, 19])
        s = pickle.loads(pickle.dumps(r))
        self.assertEqual(r, s)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

try:
    import yaml
except ImportError:
    yaml = None

import math
import unittest

import numpy as np

from lsst.sphgeom import CONTAINS, DISJOINT, Angle, Circle, Region, UnitVector3d


class CircleTestCase(unittest.TestCase):
    """Test Circle."""

    def setUp(self):
        np.random.seed(1)

    def test_construction(self):
        self.assertTrue(Circle.empty().isEmpty())
        self.assertTrue(Circle().isEmpty())
        self.assertTrue(Circle.full().isFull())
        c = Circle(UnitVector3d.X())
        self.assertEqual(c.getOpeningAngle(), Angle(0))
        self.assertEqual(c.getSquaredChordLength(), 0)
        c = Circle(UnitVector3d.Z(), 2.0)
        self.assertTrue(c.contains(UnitVector3d.Z()))
        c = Circle(UnitVector3d.Z(), Angle(math.pi))
        self.assertTrue(c.isFull())
        d = c.clone()
        self.assertEqual(c, d)
        self.assertNotEqual(id(c), id(d))
        e = Circle(d)
        self.assertEqual(d, e)

    def test_comparison_operators(self):
        c = Circle(UnitVector3d.X(), 4.0)
        d = Circle(UnitVector3d.Y(), 4.0)
        self.assertEqual(c, d)
        self.assertTrue(c.isFull())
        self.assertNotEqual(c, Circle(UnitVector3d.Z()))

    def test_center_and_dimensions(self):
        c = Circle(UnitVector3d.X(), 1)
        self.assertEqual(c.getCenter(), UnitVector3d.X())
        self.assertEqual(c.getSquaredChordLength(), 1)
        self.assertAlmostEqual(c.getOpeningAngle().asRadians(), math.pi / 3)

    def test_relationships(self):
        c = Circle(UnitVector3d.X(), Angle.fromDegrees(0.1))
        d = Circle(UnitVector3d(1, 1, 1), Angle(math.pi / 2))
        e = Circle(-UnitVector3d.X())
        self.assertTrue(c.contains(UnitVector3d.X()))
        self.assertTrue(UnitVector3d.X() in c)
        self.assertTrue(d.contains(c))
        self.assertTrue(c.isWithin(d))
        self.assertTrue(c.intersects(d))
        self.assertTrue(c.intersects(UnitVector3d.X()))
        self.assertTrue(e.isDisjointFrom(d))
        self.assertEqual(d.relate(c), CONTAINS)
        self.assertEqual(d.overlaps(c), True)
        self.assertEqual(e.relate(d), DISJOINT)
        self.assertEqual(e.overlaps(d), False)

    def test_vectorized_contains(self):
        b = Circle(UnitVector3d(*np.random.randn(3)), Angle(0.4 * math.pi))
        x = np.random.rand(5, 3)
        y = np.random.rand(5, 3)
        z = np.random.rand(5, 3)
        c = b.contains(x, y, z)
        lon = np.arctan2(y, x)
        lat = np.arctan2(z, np.hypot(x, y))
        c2 = b.contains(lon, lat)
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c[i, j], b.contains(u))
                self.assertEqual(c2[i, j], b.contains(u))
        # test with non-contiguous memory
        c3 = b.contains(x[::2], y[::2], z[::2])
        c4 = b.contains(lon[::2], lat[::2])
        for i in range(x.shape[0], 2):
            for j in range(x.shape[1]):
                u = UnitVector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c3[i // 2, j], b.contains(u))
                self.assertEqual(c4[i // 2, j], b.contains(u))

    def test_expanding_and_clipping(self):
        a = Circle.empty()
        b = (
            a.expandedTo(UnitVector3d.X())
            .expandedTo(Circle(UnitVector3d.Y(), 1))
            .clippedTo(Circle(UnitVector3d(1, 1, 0), 1))
            .clippedTo(UnitVector3d.Y())
        )
        a.expandTo(UnitVector3d.X())
        a.expandTo(Circle(UnitVector3d.Y(), 1))
        a.clipTo(Circle(UnitVector3d(1, 1, 0), 1))
        a.clipTo(UnitVector3d.Y())
        self.assertEqual(a, b)
        self.assertEqual(a, Circle(UnitVector3d.Y()))
        a.clipTo(UnitVector3d.Z())
        self.assertTrue(a.isEmpty())

    def test_dilation_and_erosion(self):
        a = Angle(math.pi / 2)
        c = Circle(UnitVector3d.X())
        d = c.dilatedBy(a).erodedBy(a)
        c.dilateBy(a).erodeBy(a)
        self.assertEqual(c, d)
        self.assertEqual(c, Circle(UnitVector3d.X()))

    def test_complement(self):
        c = Circle(UnitVector3d.X(), 2.0)
        d = c.complemented()
        c.complement()
        self.assertEqual(c, d)
        self.assertEqual(c.getCenter(), -UnitVector3d.X())
        self.assertEqual(c.getSquaredChordLength(), 2.0)

    def test_area(self):
        c = Circle(UnitVector3d(1, 1, 1), 2.0)
        self.assertAlmostEqual(c.getArea(), 2 * math.pi)

    def test_codec(self):
        c = Circle(UnitVector3d.Y(), 1.0)
        s = c.encode()
        self.assertEqual(Circle.decode(s), c)
        self.assertEqual(Region.decode(s), c)

    def test_string(self):
        c = Circle(UnitVector3d.Z(), Angle(1.0))
        self.assertEqual(str(c), "Circle([0.0, 0.0, 1.0], 1.0)")
        self.assertEqual(repr(c), "Circle(UnitVector3d(0.0, 0.0, 1.0), Angle(1.0))")
        self.assertEqual(c, eval(repr(c), {"Angle": Angle, "Circle": Circle, "UnitVector3d": UnitVector3d}))

    def test_pickle(self):
        a = Circle(UnitVector3d(1, -1, 1), 1.0)
        b = pickle.loads(pickle.dumps(a, pickle.HIGHEST_PROTOCOL))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def test_yaml(self):
        a = Circle(UnitVector3d(1, -1, 1), 1.0)
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

import hpgeom as hpg
import numpy as np

try:
    import yaml
except ImportError:
    yaml = None

from lsst.sphgeom import Angle, Box, Circle, ConvexPolygon, Ellipse, HealpixPixelization, LonLat, UnitVector3d


class HealpixPixelizationTestCase(unittest.TestCase):
    """Test HEALPix pixelization."""

    def test_construction(self):
        """Test construction of a HealpixPixelization."""
        with self.assertRaises(ValueError):
            HealpixPixelization(-1)
        h1 = HealpixPixelization(5)
        self.assertEqual(h1.level, 5)
        self.assertEqual(h1.getLevel(), 5)
        self.assertEqual(h1.nside, 32)
        h2 = HealpixPixelization(6)
        h3 = HealpixPixelization(h2.level)
        self.assertNotEqual(h1, h2)
        self.assertEqual(h2, h3)

    def test_indexing(self):
        """Test indexing of HealpixPixelization."""
        h = HealpixPixelization(5)
        vec = UnitVector3d(1, 1, 1)
        lonlat = LonLat(vec)
        pix = hpg.angle_to_pixel(h.nside, lonlat.getLon().asDegrees(), lonlat.getLat().asDegrees())
        self.assertEqual(h.index(UnitVector3d(1, 1, 1)), pix)

    def test_pixel(self):
        """Test pixel polygon of HealpixPixelization."""
        h = HealpixPixelization(5)
        pix_poly = h.pixel(10)
        self.assertIsInstance(pix_poly, ConvexPolygon)

    def test_envelope(self):
        """Test envelope method of HealpixPixelization."""
        # Make the hardest intersection: a region that _just_
        # touches a healpix pixel.
        h = HealpixPixelization(5)
        pix = hpg.angle_to_pixel(h.nside, 50.0, 20.0)

        corners_ra, corners_dec = hpg.boundaries(h.nside, pix, step=1)

        # Take the southernmost corner...
        smost = np.argmin(corners_dec)

        # Choose challenging comparison box corners:
        ra_range = np.array([corners_ra[smost] - 0.5, corners_ra[smost] + 0.5])
        dec_range = np.array([corners_dec[smost] - 0.5, corners_dec[smost] + 1e-8])

        # Test the box region
        box = Box(
            point1=LonLat.fromDegrees(ra_range[0], dec_range[0]),
            point2=LonLat.fromDegrees(ra_range[1], dec_range[1]),
        )
        # These pixels have been checked to completely overlap the region
        self._check_envelope(h, box, [98, 99, 104, 105])

        # Try a polygon region:
        poly = ConvexPolygon(
            [
                UnitVector3d(LonLat.fromDegrees(ra_range[0], dec_range[0])),
                UnitVector3d(LonLat.fromDegrees(ra_range[1], dec_range[0])),
                UnitVector3d(LonLat.fromDegrees(ra_range[1], dec_range[1])),
                UnitVector3d(LonLat.fromDegrees(ra_range[0], dec_range[1])),
                UnitVector3d(LonLat.fromDegrees(ra_range[0], dec_range[0])),
            ]
        )
        self._check_envelope(h, poly, [98, 99, 104, 105])

        # Try a circle region
        circle = Circle(
            center=UnitVector3d(
                LonLat.fromDegrees((ra_range[0] + ra_range[1]) / 2.0, (dec_range[0] + dec_range[1]) / 2.0)
            ),
            angle=Angle.fromDegrees((dec_range[1] - dec_range[0]) / 2.0),
        )
        self._check_envelope(h, circle, [98, 99, 104, 105])

        # Try a circle region with zero-radius (representing a point)
        pt_circle = Circle(
            center=UnitVector3d(
                LonLat.fromDegrees((ra_range[0] + ra_range[1]) / 2.0, (dec_range[0] + dec_range[1]) / 2.0)
            ),
        )
        self._check_envelope(h, pt_circle, [98])

    def test_envelope_missing_neighbors(self):
        """Test envelope, with a pixel with missing neighbors.

        Testing DM-47043.
        """
        h = HealpixPixelization(0)
        h_highres = HealpixPixelization(4)

        self.assertEqual(h._nside_highres, h_highres._nside)

        # Find a pixel with a missing neighbor at high res.
        n = hpg.neighbors(h_highres._nside, np.arange(hpg.nside_to_npixel(h_highres._nside)))
        ind = np.argmin(n)
        self.assertEqual(n.ravel()[ind], -1)

        # This is the pixel with a bad (-1) neighbor.
        badpix = ind // 8
        ra, dec = hpg.pixel_to_angle(h_highres._nside, badpix)

        box = Box(
            point1=LonLat.fromDegrees(ra - 0.1, dec - 0.1),
            point2=LonLat.fromDegrees(ra + 0.1, dec + 0.1),
        )

        self._check_envelope(h, box, [0, 1, 5])

    def _check_envelope(self, pixelization, region, check_pixels):
        """Check the envelope from a region.

        Parameters
        ----------
        pixelization : `lsst.sphgeom.HealpixPixelization`
        region : `lsst.sphgeom.Region`
        check_pixels : `list` [`int`]
        """
        pixel_range = pixelization.envelope(region)

        pixels = []
        for r in pixel_range.ranges():
            pixels.extend(range(r[0], r[1]))

        self.assertEqual(pixels, check_pixels)

    def test_interior(self):
        """Test interior method of HealpixPixelization."""
        h = HealpixPixelization(5)
        pix = hpg.angle_to_pixel(h.nside, 50.0, 20.0)

        corners_ra, corners_dec = hpg.boundaries(h.nside, pix, step=1)

        ra_range = np.array([corners_ra.min() - 1.0, corners_ra.max() + 1.0])
        dec_range = np.array([corners_dec.min() - 1.0, corners_dec.max() + 1.0])

        # Test the box region
        box = Box(
            point1=LonLat.fromDegrees(ra_range[0], dec_range[0]),
            point2=LonLat.fromDegrees(ra_range[1], dec_range[1]),
        )
        # These pixels have been checked to completely overlap the region
        self._check_interior(h, box, [pix])

        # Try a polygon region:
        poly = ConvexPolygon(
            [
                UnitVector3d(LonLat.fromDegrees(ra_range[0], dec_range[0])),
                UnitVector3d(LonLat.fromDegrees(ra_range[1], dec_range[0])),
                UnitVector3d(LonLat.fromDegrees(ra_range[1], dec_range[1])),
                UnitVector3d(LonLat.fromDegrees(ra_range[0], dec_range[1])),
                UnitVector3d(LonLat.fromDegrees(ra_range[0], dec_range[0])),
            ]
        )
        self._check_interior(h, poly, [pix])

        # Try a circle region
        circle = Circle(
            center=UnitVector3d(
                LonLat.fromDegrees((ra_range[0] + ra_range[1]) / 2.0, (dec_range[0] + dec_range[1]) / 2.0)
            ),
            angle=Angle.fromDegrees(2.5),
        )
        self._check_interior(h, circle, [pix])

        # Try an ellipse region
        ellipse = Ellipse(
            center=UnitVector3d(
                LonLat.fromDegrees((ra_range[0] + ra_range[1]) / 2.0, (dec_range[0] + dec_range[1]) / 2.0)
            ),
            alpha=Angle.fromDegrees(1.5),
            beta=Angle.fromDegrees(2.5),
            orientation=Angle.fromDegrees(45.0),
        )
        self._check_interior(h, ellipse, [pix])

    def test_interior_nearly_degenerate_polygon(self):
        """Check interior pixels with nearly degenerate polygon.

        See DM-53933.
        """
        poly = ConvexPolygon(
            [
                UnitVector3d(-0.016144196513206758, -0.9235039286924489, -0.3832490816799893),
                UnitVector3d(-0.01988138785117867, -0.9228560783513854, -0.38463149775728545),
                UnitVector3d(-0.022348863683228713, -0.9209177038528237, -0.3891158066983548),
                UnitVector3d(-0.024841110338799998, -0.918545770328335, -0.3945333788782152),
                UnitVector3d(-0.026637426968850634, -0.9164294757774608, -0.39930873194901123),
                UnitVector3d(-0.02509225617228684, -0.9149335621264763, -0.4028237276709774),
                UnitVector3d(-0.0221015078570923, -0.911998223315482, -0.4095982959191202),
                UnitVector3d(-0.020553220684181986, -0.9104571830599119, -0.4130923418972051),
                UnitVector3d(-0.017371565494301272, -0.9072522623419655, -0.4202279871541906),
                UnitVector3d(-0.014347349160349979, -0.9041551644174409, -0.4269632211670476),
                UnitVector3d(-0.014303105070771129, -0.9041094310633312, -0.42706153871271785),
                UnitVector3d(-0.012802279210872009, -0.9025514012895127, -0.4303917630221838),
                UnitVector3d(-0.007862275539177377, -0.9015265103248066, -0.43265244227315125),
                UnitVector3d(-0.0019484762556141095, -0.900471339489277, -0.4349109911219402),
                UnitVector3d(0.0032282853511799436, -0.8996943229589627, -0.436508537613075),
                UnitVector3d(0.006969847869770022, -0.9003377930402403, -0.4351359323752772),
                UnitVector3d(0.022099717156107892, -0.9027944227880381, -0.42950417074160424),
                UnitVector3d(0.033589177231523826, -0.9045049275054283, -0.4251383342999184),
                UnitVector3d(0.03732521827078148, -0.9050320420971739, -0.4237025263772424),
                UnitVector3d(0.039795546158565454, -0.9070052460628258, -0.41923477684998206),
                UnitVector3d(0.04228779372661228, -0.9093772751891867, -0.41381724694752126),
                UnitVector3d(0.04409395578698435, -0.9114581951867685, -0.4090228373696683),
                UnitVector3d(0.04255005806475871, -0.9130743444969162, -0.4055671756691021),
                UnitVector3d(0.04109145924730725, -0.91458405813035, -0.4023027374885082),
                UnitVector3d(0.03791490987091337, -0.9178181221381234, -0.3951864803916365),
                UnitVector3d(0.03335712714052637, -0.9223387473238281, -0.38493965404208763),
                UnitVector3d(0.03180762744041508, -0.9238417869862389, -0.3814506881035668),
                UnitVector3d(0.030248610691484885, -0.9253347742243898, -0.3779425580195124),
                UnitVector3d(0.02530895801451361, -0.9263964395908428, -0.3756981412751856),
                UnitVector3d(0.01939515964342646, -0.9274517060282444, -0.37343963470379626),
                UnitVector3d(0.014215536919345694, -0.9281937989444676, -0.37182548340738003),
                UnitVector3d(0.010475126504663994, -0.9276710642902235, -0.3732514811803906),
            ],
        )

        h = HealpixPixelization(6)
        self._check_interior(h, poly, [28890, 28891, 28912, 28913, 28914])

    def _check_interior(self, pixelization, region, check_pixels):
        """Check the interior from a region.

        Parameters
        ----------
        pixelization : `lsst.sphgeom.HealpixPixelization`
        region : `lsst.sphgeom.Region`
        check_pixels : `list` [`int`]
        """
        pixel_range = pixelization.interior(region)

        pixels = []
        for r in pixel_range.ranges():
            pixels.extend(range(r[0], r[1]))

        self.assertEqual(pixels, check_pixels)

    def test_index_to_string(self):
        """Test converting index to string of HealpixPixelization."""
        h = HealpixPixelization(5)
        self.assertEqual(h.toString(0), str(0))
        self.assertEqual(h.toString(100), str(100))

    def test_string(self):
        """Test string representation of HealpixPixelization."""
        h = HealpixPixelization(5)
        self.assertEqual(str(h), "HealpixPixelization(5)")
        self.assertEqual(str(h), repr(h))
        self.assertEqual(h, eval(repr(h), {"HealpixPixelization": HealpixPixelization}))

    def test_pickle(self):
        """Test pickling of HealpixPixelization."""
        a = HealpixPixelization(5)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def test_yaml(self):
        """Test yaml representation of HealpixPixelization."""
        a = HealpixPixelization(5)
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

import numpy as np

from lsst.sphgeom import CONTAINS, DISJOINT, Box3d, Interval1d, Vector3d


class Box3dTestCase(unittest.TestCase):
    """Test Box3d."""

    def setUp(self):
        np.random.seed(1)

    def test_construction(self):
        a = Box3d(Vector3d(0, 0, 0))
        b = Box3d(a)
        self.assertEqual(a, b)
        self.assertNotEqual(id(a), id(b))
        a = Box3d(Vector3d(1, 2, 3), Vector3d(3, 4, 5))
        b = Box3d(Interval1d(1, 3), Interval1d(2, 4), Interval1d(3, 5))
        c = Box3d(Vector3d(2, 3, 4), 1, 1, 1)
        self.assertEqual(a, b)
        self.assertEqual(b, c)
        i = Interval1d(1, 2)
        self.assertEqual(i, Interval1d(1, 2))
        self.assertTrue(Interval1d.empty().isEmpty())

    def test_comparison_operators(self):
        self.assertEqual(Box3d(Vector3d(1, 1, 1)), Vector3d(1, 1, 1))
        self.assertEqual(Box3d(Vector3d(1, 1, 1)), Box3d(Vector3d(1, 1, 1), Vector3d(1, 1, 1)))
        self.assertEqual(Box3d(Vector3d(0, 0, 0), 1, 1, 1), Box3d(Vector3d(-1, -1, -1), Vector3d(1, 1, 1)))
        self.assertNotEqual(Box3d(Vector3d(0, 0, 0), 1, 1, 1), Box3d(Vector3d(-1, -1, -1), Vector3d(1, 1, 2)))
        self.assertNotEqual(Box3d(Vector3d(0, 0, 0), 1, 1, 1), Vector3d(1, 1, 1))

    def test_center_and_dimensions(self):
        b = Box3d(Vector3d(1.5, 1.5, 1.5), 0.5, 1.0, 1.5)
        self.assertEqual(b[0], b[-3])
        self.assertEqual(b[1], b[-2])
        self.assertEqual(b[2], b[-1])
        self.assertEqual(b[0], b.x())
        self.assertEqual(b[1], b.y())
        self.assertEqual(b[2], b.z())
        with self.assertRaises(IndexError):
            b[-4]
        with self.assertRaises(IndexError):
            b[3]
        self.assertEqual(b.x(), Interval1d(1, 2))
        self.assertEqual(b.y(), Interval1d(0.5, 2.5))
        self.assertEqual(b.z(), Interval1d(0, 3))
        self.assertEqual(b.getCenter(), Vector3d(1.5, 1.5, 1.5))
        self.assertEqual(b.getWidth(), 1)
        self.assertEqual(b.getHeight(), 2)
        self.assertEqual(b.getDepth(), 3)
        self.assertEqual(b.isEmpty(), False)
        self.assertEqual(b.isFull(), False)

    def test_relationships(self):
        b02 = Box3d(Interval1d(0, 2), Interval1d(0, 2), Interval1d(0, 2))
        b13 = Box3d(Interval1d(1, 3), Interval1d(1, 3), Interval1d(1, 3))
        b46 = Box3d(Interval1d(4, 6), Interval1d(4, 6), Interval1d(4, 6))
        b06 = Box3d(Interval1d(0, 6), Interval1d(0, 6), Interval1d(0, 6))
        self.assertTrue(b02.contains(Vector3d(1, 1, 1)))
        self.assertTrue(b02.contains(Box3d(Vector3d(1, 1, 1), 0.5, 0.5, 0.5)))
        self.assertTrue(b02.isDisjointFrom(Vector3d(3, 3, 3)))
        self.assertTrue(b02.isDisjointFrom(b46))
        self.assertTrue(b02.intersects(Vector3d(1, 1, 1)))
        self.assertTrue(b02.intersects(b13))
        self.assertTrue(Box3d(Vector3d(1, 1, 1), 0, 0, 0).isWithin(b02))
        self.assertTrue(b02.isWithin(b06))
        r = b02.relate(Vector3d(1, 1, 1))
        self.assertEqual(r, CONTAINS)
        r = b46.relate(b02)
        self.assertEqual(r, DISJOINT)

    def test_vectorized_contains(self):
        b = Box3d.aroundUnitSphere()
        x = np.random.rand(5, 3)
        y = np.random.rand(5, 3)
        z = np.random.rand(5, 3)
        c = b.contains(x, y, z)
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                u = Vector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c[i, j], b.contains(u))
        # test with non-contiguous memory
        c2 = b.contains(x[::2], y[::2], z[::2])
        for i in range(x.shape[0], 2):
            for j in range(x.shape[1]):
                u = Vector3d(x[i, j], y[i, j], z[i, j])
                self.assertEqual(c2[i // 2, j], b.contains(u))

    def test_expanding_and_clipping(self):
        a = Box3d(Vector3d(1, 1, 1), Vector3d(2, 2, 2))
        b = (
            a.expandedTo(Vector3d(3, 3, 3))
            .expandedTo(Box3d(Vector3d(3, 3, 3), 1, 1, 1))
            .clippedTo(Box3d(Vector3d(1, 1, 1), 1, 1, 1))
            .clippedTo(Vector3d(1, 1, 1))
        )
        a.expandTo(Vector3d(3, 3, 3))
        a.expandTo(Box3d(Vector3d(3, 3, 3), 1, 1, 1))
        a.clipTo(Box3d(Vector3d(1, 1, 1), 1, 1, 1))
        a.clipTo(Vector3d(1, 1, 1))
        self.assertEqual(a, b)
        self.assertEqual(a, Vector3d(1, 1, 1))
        a.clipTo(Vector3d(0, 0, 0))
        self.assertTrue(a.isEmpty())

    def test_dilation_and_erosion(self):
        a = Box3d(Vector3d(0, 0, 0), 1, 1, 1)
        b = a.dilatedBy(1).erodedBy(2)
        a.dilateBy(1).erodeBy(2)
        self.assertEqual(a, b)
        self.assertEqual(a, Vector3d(0, 0, 0))

    def test_string(self):
        b = Box3d(Vector3d(0, 0, 0), 1, 1, 1)
        self.assertEqual(str(b), "[[-1.0, 1.0],\n [-1.0, 1.0],\n [-1.0, 1.0]]")
        self.assertEqual(
            repr(b),
            "Box3d(Interval1d(-1.0, 1.0),\n      Interval1d(-1.0, 1.0),\n      Interval1d(-1.0, 1.0))",
        )
        self.assertEqual(b, eval(repr(b), {"Box3d": Box3d, "Interval1d": Interval1d}))

    def test_pickle(self):
        a = Box3d(Vector3d(0, 0, 0), 1, 1, 1)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import Angle, LonLat, NormalizedAngle, UnitVector3d


class LonLatTestCase(unittest.TestCase):
    """Test LonLat."""

    def testConstruction(self):
        p = LonLat.fromDegrees(45, 45)
        self.assertEqual(p, LonLat(NormalizedAngle.fromDegrees(45), Angle.fromDegrees(45)))
        u = UnitVector3d(p)
        q = LonLat(u)
        self.assertAlmostEqual(p.getLon().asRadians(), q.getLon().asRadians(), places=13)
        self.assertAlmostEqual(p.getLat().asRadians(), q.getLat().asRadians(), places=13)
        self.assertAlmostEqual(p.getLon().asRadians(), LonLat.latitudeOf(u).asRadians(), places=13)
        self.assertAlmostEqual(p.getLon().asRadians(), LonLat.longitudeOf(u).asRadians(), places=13)

    def testComparisonOperators(self):
        self.assertEqual(LonLat.fromDegrees(45, 45), LonLat.fromDegrees(45, 45))
        self.assertNotEqual(LonLat.fromDegrees(0, 0), LonLat.fromDegrees(45, 45))

    def testString(self):
        p = LonLat.fromRadians(1, 1)
        self.assertEqual(str(p), "[1.0, 1.0]")
        self.assertEqual(repr(p), "LonLat.fromRadians(1.0, 1.0)")
        self.assertEqual(p, eval(repr(p), {"LonLat": LonLat}))

    def testPickle(self):
        p = LonLat.fromRadians(2, 1)
        q = pickle.loads(pickle.dumps(p))
        self.assertEqual(p, q)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
__all__ = ["PixelizationABC"]

import abc

from ._sphgeom import RangeSet, Region, UnitVector3d


class PixelizationABC(abc.ABC):
    """Pixelization ABC class that should be a base for
    Python implementations of pixelization.
    """

    @abc.abstractmethod
    def universe(self) -> RangeSet:
        """Return the set of all pixel indexes for this pixelization.

        Returns
        -------
        rangeSet : `lsst.sphgeom.RangeSet`
        """
        pass

    @abc.abstractmethod
    def pixel(self, i) -> Region:
        """Return the spherical region corresponding to the pixel index ``i``.

        This region will contain all unit vectors v with ``index(v) == i``.
        But it may also contain points with index not equal to ``i``.
        To see why, consider a point that lies on the edge of a polygonal
        pixel - it is inside the polygons for both pixels sharing the edge,
        but must be assigned to exactly one pixel by the pixelization.

        Parameters
        ----------
        i : `int`
            Pixel index.

        Returns
        -------
        region : `lsst.sphgeom.Region`
            The spherical region corresponding to the pixel with index ``i``

        Raises
        ------
        `InvalidArgumentException`
            Raised if ``i`` is not a valid pixel index.
        """
        pass

    @abc.abstractmethod
    def index(self, v: UnitVector3d) -> int:
        """Compute the index of the pixel.

        Parameters
        ----------
        v : `lsst.sphgeom.UnitVector3d`

        Returns
        -------
        i : `int`
            The index of the pixel.
        """
        pass

    @abc.abstractmethod
    def toString(self, i: int) -> str:
        """Convert the given pixel index to a human-readable string.

        Parameters
        ----------
        i : `int`

        Returns
        -------
        s : `str`
        """
        pass

    @abc.abstractmethod
    def envelope(self, region: Region, maxRanges: int = 0):
        """Return the indexes of the pixels intersecting the spherical region.

        The ``maxRanges`` parameter can be used to limit both these costs -
        setting it to a non-zero value sets a cap on the number of ranges
        returned by this method. To meet this constraint, implementations are
        allowed to return pixels that do not intersect the region along with
        those, that do.
        This allows two ranges [a, b) and [c, d), a < b < c < d, to be
        merged into one range [a, d) (by adding in the pixels [b, c)). Since
        simplification proceeds by adding pixels, the return value will always
        be a superset of the intersecting pixels.

        Parameters
        ----------
        region : `lsst.sphgeom.Region`
        maxRanges : `int`

        Returns
        -------
        rangeSet : `lsst.sphgeom.RangeSet`
        """
        pass

    @abc.abstractmethod
    def interior(self, region: Region, maxRanges: int = 0):
        """Return the indexes of the pixels within the spherical region.

        The ``maxRanges`` argument is analogous to the identically named
        envelope() argument. The only difference is that implementations must
        remove interior pixels to keep the number of ranges at or below the
        maximum. The return value is therefore always a subset of the interior
        pixels.

        Parameters
        ----------
        region : `lsst.sphgeom.Region`
        maxRanges : `int`

        Returns
        -------
        rangeSet : `lsst.sphgeom.RangeSet`
        """
        pass



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
__all__ = ["HealpixPixelization"]

import hpgeom as hpg
import numpy as np

from ._sphgeom import Box, Circle, ConvexPolygon, Ellipse, LonLat, RangeSet, Region, UnitVector3d
from .pixelization_abc import PixelizationABC


class HealpixPixelization(PixelizationABC):
    """HEALPix pixelization class.

    Parameters
    ----------
    level : `int`
        Pixelization level.  HEALPix nside = 2**level
    """

    MAX_LEVEL = 17

    def __init__(self, level: int):
        if level < 0:
            raise ValueError("HealPix level must be >= 0.")

        self._level = level
        self._nside = 2**level

        self._npix = hpg.nside_to_npixel(self._nside)

        # Values used to do pixel/region intersections
        self._bit_shift = 8
        self._nside_highres = self._nside * (2 ** (self._bit_shift // 2))

    @property
    def nside(self):
        return self._nside

    def getLevel(self):
        return self._level

    level = property(getLevel)

    def universe(self) -> RangeSet:
        return RangeSet(0, self._npix)

    def pixel(self, i) -> Region:
        # This is arbitrarily returning 4 points on a side
        # to approximate the pixel shape.
        varr = hpg.angle_to_vector(*hpg.boundaries(self._nside, i, step=4))
        return ConvexPolygon([UnitVector3d(*c) for c in varr])

    def index(self, v: UnitVector3d) -> int:
        return hpg.vector_to_pixel(self._nside, v.x(), v.y(), v.z())

    def toString(self, i: int) -> str:
        return str(i)

    def envelope(self, region: Region, maxRanges: int = 0):
        if maxRanges > 0:
            # If this is important, the rangeset can be consolidated.
            raise NotImplementedError("HealpixPixelization: maxRanges not implemented")
        pixels_highres = self._interior_pixels_from_region(self._nside_highres, region)

        # Dilate the high resolution pixels by one to ensure that the full
        # region is completely covered at high resolution.
        neighbors = hpg.neighbors(self._nside_highres, pixels_highres).ravel()
        # Shift back to the original resolution and uniquify
        pixels = np.unique(np.right_shift(neighbors[neighbors >= 0], self._bit_shift))

        return RangeSet(pixels)

    def interior(self, region: Region, maxRanges: int = 0):
        if maxRanges > 0:
            # If this is important, the rangeset can be consolidated.
            raise NotImplementedError("HealpixPixelization: maxRanges not implemented")
        pixels = self._interior_pixels_from_region(self._nside, region)

        # Check that the corners of the pixels are entirely enclosed in
        # the region

        # Returns arrays [npixels, ncorners], where ncorners is 4.
        corners_lon, corners_lat = hpg.boundaries(self._nside, pixels, step=1, degrees=False)

        corners_int = region.contains(corners_lon.ravel(), corners_lat.ravel()).reshape((len(pixels), 4))
        interior = np.sum(corners_int, axis=1) == 4
        pixels = pixels[interior]

        return RangeSet(pixels)

    def _interior_pixels_from_region(self, nside: int, region: Region):
        """Get interior pixels from a region.

        Parameters
        ----------
        nside : `int`
            Healpix nside to retrieve interior pixels.
        region : `lsst.sphgeom.Region`
            Sphgeom region to find interior pixels.

        Returns
        -------
        pixels : `np.ndarray`
            Array of pixels at resolution nside, nest ordering.
        """
        if isinstance(region, Circle):
            center = LonLat(region.getCenter())
            radius_rad = region.getOpeningAngle().asRadians()
            if radius_rad > 0:
                pixels = hpg.query_circle(
                    nside,
                    center.getLon().asRadians(),
                    center.getLat().asRadians(),
                    radius_rad,
                    degrees=False,
                )
            elif radius_rad == 0:
                pixels = hpg.angle_to_pixel(
                    nside,
                    center.getLon().asRadians(),
                    center.getLat().asRadians(),
                    degrees=False,
                )
            else:
                raise ValueError("Invalid circle radius.")
        elif isinstance(region, ConvexPolygon):
            vertices = np.array([[v.x(), v.y(), v.z()] for v in region.getVertices()])
            retry = False
            try:
                pixels = hpg.query_polygon_vec(nside, vertices)
            except RuntimeError:
                # This happens if the polygon is determined to have a
                # degenerate corner (which can happen with rounding).
                # In this case, redo with a bounding circle.
                retry = True

            if retry:
                pixels = self._interior_pixels_from_region(nside, region.getBoundingCircle())
        elif isinstance(region, Box):
            pixels = hpg.query_box(
                nside,
                region.getLon().getA().asRadians(),
                region.getLon().getB().asRadians(),
                region.getLat().getA().asRadians(),
                region.getLat().getB().asRadians(),
                degrees=False,
            )
        elif isinstance(region, Ellipse):
            # hpgeom supports query_ellipse given center, alpha, beta,
            # and orientation. However, until we figure out how to get
            # the orientation out of the Ellipse region, we will use the
            # bounding circle as was done with healpy.
            _circle = region.getBoundingCircle()
            center = LonLat(_circle.getCenter())
            pixels = hpg.query_circle(
                nside,
                center.getLon().asRadians(),
                center.getLat().asRadians(),
                _circle.getOpeningAngle().asRadians(),
                degrees=False,
            )
        else:
            raise ValueError("Invalid region.")

        return pixels

    def __eq__(self, other):
        if isinstance(other, HealpixPixelization):
            return self._level == other._level

    def __repr__(self):
        return f"HealpixPixelization({self._level})"

    def __reduce__(self):
        return (self.__class__, (self._level,))



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import Angle, LonLat, NormalizedAngle, UnitVector3d


class NormalizedAngleTestCase(unittest.TestCase):
    """Test normalized angle."""

    def testConstruction(self):
        a1 = NormalizedAngle(1.0)
        a2 = NormalizedAngle.fromRadians(1.0)
        a3 = NormalizedAngle.fromDegrees(57.29577951308232)
        self.assertEqual(a1, a2)
        self.assertEqual(a1.asRadians(), 1.0)
        self.assertEqual(a1, a3)
        self.assertEqual(a1.asDegrees(), 57.29577951308232)
        self.assertEqual(NormalizedAngle.between(NormalizedAngle(0), NormalizedAngle(1)), NormalizedAngle(1))
        a = NormalizedAngle.center(NormalizedAngle(0), NormalizedAngle(1))
        self.assertAlmostEqual(a.asRadians(), 0.5, places=15)
        a = NormalizedAngle(LonLat.fromDegrees(45, 0), LonLat.fromDegrees(90, 0))
        self.assertAlmostEqual(a.asDegrees(), 45.0, places=13)
        a = NormalizedAngle(UnitVector3d.Y(), UnitVector3d.Z())
        self.assertAlmostEqual(a.asDegrees(), 90.0, places=13)

    def testComparisonOperators(self):
        a1 = NormalizedAngle(1)
        a2 = NormalizedAngle(2)
        self.assertNotEqual(a1, a2)
        self.assertLess(a1, a2)
        self.assertLessEqual(a1, a2)
        self.assertGreater(a2, a1)
        self.assertGreaterEqual(a2, a1)

    def testArithmeticOperators(self):
        a = NormalizedAngle(1)
        b = -a
        self.assertEqual(a + b, Angle(0))
        self.assertEqual(a - b, 2.0 * a)
        self.assertEqual(a - b, a * 2.0)
        self.assertEqual(a / 1.0, a)
        self.assertEqual(a / a, 1.0)

    def testAngleTo(self):
        self.assertEqual(NormalizedAngle(1).getAngleTo(NormalizedAngle(2)), NormalizedAngle(1))

    def testString(self):
        self.assertEqual(str(NormalizedAngle(1)), "1.0")
        self.assertEqual(repr(NormalizedAngle(1)), "NormalizedAngle(1.0)")
        a = NormalizedAngle(0.5)
        self.assertEqual(a, eval(repr(a), {"NormalizedAngle": NormalizedAngle}))

    def testPickle(self):
        a = NormalizedAngle(1.5)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import Box, Chunker


class ChunkerTestCase(unittest.TestCase):
    """Test Chunker."""

    def testConstruction(self):
        chunker = Chunker(85, 12)
        self.assertEqual(chunker.numStripes, 85)
        self.assertEqual(chunker.numSubStripesPerStripe, 12)

    def testComparisonOperators(self):
        c = Chunker(85, 12)
        self.assertEqual(c, c)
        self.assertEqual(c, Chunker(85, 12))
        self.assertNotEqual(c, Chunker(85, 10))

    def testIntersecting(self):
        b = Box.fromDegrees(273.6, 30.7, 273.7180105379097, 30.722546655347717)
        c = Chunker(85, 12)
        self.assertEqual(c.getChunksIntersecting(b), [9630, 9631, 9797])
        self.assertEqual(c.getSubChunksIntersecting(b), [(9630, [770]), (9631, [759]), (9797, [11])])

    def testString(self):
        chunker = Chunker(85, 12)
        self.assertEqual(str(chunker), "Chunker(85, 12)")
        self.assertEqual(repr(chunker), "Chunker(85, 12)")
        self.assertEqual(chunker, eval(repr(chunker), {"Chunker": Chunker}))

    def testPickle(self):
        a = Chunker(85, 12)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)

    def testChunkBoundingBox(self):
        chunker = Chunker(200, 5)
        chunk_id = 3645
        stripe = chunker.getStripe(chunk_id)
        chunk_in_stripe = chunker.getChunk(chunk_id, stripe)
        bbox = chunker.getChunkBoundingBox(stripe, chunk_in_stripe)
        sbbox = chunker.getSubChunkBoundingBox(0, 0)
        self.assertEqual(stripe, 9)
        self.assertEqual(chunk_in_stripe, 45)
        b = Box.fromRadians(5.048988193233824, -1.4294246573883558, 5.1611879309330035, -1.413716694110407)
        self.assertAlmostEqual(bbox, b)
        sb = Box.fromRadians(0.0, -1.5707963267948966, 6.283185307179586, -1.5676547341363067)
        self.assertAlmostEqual(sbbox, sb)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle

try:
    import yaml
except ImportError:
    yaml = None

import unittest

from lsst.sphgeom import (
    Angle,
    Circle,
    ConvexPolygon,
    HtmPixelization,
    IntersectionRegion,
    RangeSet,
    UnionRegion,
    UnitVector3d,
)


class HtmPixelizationTestCase(unittest.TestCase):
    """Test HTM pixels."""

    def test_construction(self):
        with self.assertRaises(ValueError):
            HtmPixelization(-1)
        with self.assertRaises(ValueError):
            HtmPixelization(HtmPixelization.MAX_LEVEL + 1)
        h1 = HtmPixelization(0)
        self.assertEqual(h1.getLevel(), 0)
        h2 = HtmPixelization(1)
        h3 = HtmPixelization(h2)
        self.assertNotEqual(h1, h2)
        self.assertEqual(h2, h3)

    def test_indexing(self):
        h = HtmPixelization(1)
        self.assertEqual(h.index(UnitVector3d(1, 1, 1)), 63)

    def test_pixel(self):
        h = HtmPixelization(1)
        self.assertIsInstance(h.pixel(10), ConvexPolygon)

    def test_level(self):
        for index in (0, 16 * 4**HtmPixelization.MAX_LEVEL):
            self.assertEqual(HtmPixelization.level(index), -1)
        for level in range(HtmPixelization.MAX_LEVEL + 1):
            for root in range(8, 16):
                self.assertEqual(HtmPixelization.level(root * 4**level), level)

    def test_envelope_and_interior(self):
        pixelization = HtmPixelization(3)
        c = Circle(UnitVector3d(1, 1, 1), Angle.fromDegrees(0.1))
        rs1 = pixelization.envelope(c)
        self.assertEqual(rs1, RangeSet(0x3FF))
        rs = pixelization.envelope(c, 1)
        self.assertEqual(rs, RangeSet(0x3FF))
        self.assertTrue(rs.isWithin(pixelization.universe()))
        rs = pixelization.interior(c)
        self.assertTrue(rs.empty())

        # Create a second region for a union.
        s3 = UnitVector3d(1.0, -1.0, 1.0)  # Center of S3
        c2 = Circle(s3, 1e-8)
        rs2 = pixelization.envelope(c2)
        self.assertEqual(rs2.ranges(), [(831, 832)])

        # Try again with a union.
        union = UnionRegion(c, c2)
        rsu = pixelization.envelope(union)
        self.assertEqual(rsu, rs1 | rs2)

        # Check that nested unions also work.
        c3 = Circle(s3, 2e-8)
        union2 = UnionRegion(union, c3)
        rsu2 = pixelization.envelope(union2)
        self.assertEqual(rsu2, rsu2 | rsu)

        # Check with intersection
        c4 = Circle(UnitVector3d(1.0, 1.0, 2.0), 1)
        c5 = Circle(UnitVector3d(1.0, 1.0, 2.5), 0.5)
        rs4 = pixelization.envelope(c4)
        rs5 = pixelization.envelope(c5)
        intersection = IntersectionRegion(c4, c5)
        rsi = pixelization.envelope(intersection)
        self.assertEqual(rsi, rs4 & rs5)

        # Check that nested intersection also work.
        c6 = Circle(UnitVector3d(1.0, 1.0, 2.0), 2)
        intersection2 = IntersectionRegion(intersection, c6)
        rsi2 = pixelization.envelope(intersection2)
        self.assertEqual(rsi2, rsi2 & rsi)

        # Check with empty union.
        union3 = UnionRegion()
        rsu3 = pixelization.envelope(union3)
        self.assertTrue(rsu3.empty())

        # Check with empty intersection, which is the same as the full sky.
        intersection3 = IntersectionRegion()
        rsi3 = pixelization.envelope(intersection3)
        self.assertEqual(rsi3, pixelization.universe())

    def test_index_to_string(self):
        strings = ["S0", "S1", "S2", "S3", "N0", "N1", "N2", "N3"]
        for i in range(8, 16):
            s0 = strings[i - 8]
            self.assertEqual(HtmPixelization.asString(i), s0)
            self.assertEqual(HtmPixelization(0).toString(i), s0)
            for j in range(4):
                s1 = s0 + str(j)
                self.assertEqual(HtmPixelization.asString(i * 4 + j), s1)
                self.assertEqual(HtmPixelization(1).asString(i * 4 + j), s1)

    def test_string(self):
        p = HtmPixelization(3)
        self.assertEqual(str(p), "HtmPixelization(3)")
        self.assertEqual(str(p), repr(p))
        self.assertEqual(p, eval(repr(p), {"HtmPixelization": HtmPixelization}))

    def test_pickle(self):
        a = HtmPixelization(20)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)

    @unittest.skipIf(not yaml, "YAML module can not be imported")
    def test_yaml(self):
        a = HtmPixelization(20)
        b = yaml.safe_load(yaml.dump(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Add support for YAML serialization of regions."""

__all__ = ()

# YAML representers and constructors are only useful if yaml itself can
# be imported.  If yaml can not be imported then there is no reason to
# break the entire module import of sphgeom.
try:
    import yaml
except ImportError:
    yaml = None

from ._healpixPixelization import HealpixPixelization
from ._sphgeom import (
    Box,
    Circle,
    ConvexPolygon,
    Ellipse,
    HtmPixelization,
    IntersectionRegion,
    Mq3cPixelization,
    Q3cPixelization,
    Region,
    UnionRegion,
)

if yaml:
    try:
        # CLoader is not always available
        from yaml import CLoader
    except ImportError:
        CLoader = None

    YamlLoaders = (yaml.Loader, yaml.FullLoader, yaml.SafeLoader, yaml.UnsafeLoader)
    if CLoader is not None:
        YamlLoaders += (CLoader,)


# Regions


def region_representer(dumper, data):
    """Represent a sphgeom region object in a form suitable for YAML.

    Stores the region as a mapping with a single ``encoded`` key
    storing the hex encoded byte string.
    """
    encoded = data.encode()
    return dumper.represent_mapping(f"lsst.sphgeom.{type(data).__name__}", {"encoded": encoded.hex()})


def region_constructor(loader, node):
    """Construct a sphgeom region from YAML."""
    mapping = loader.construct_mapping(node)
    encoded = bytes.fromhex(mapping["encoded"])
    # The generic Region base class can instantiate a region of the
    # correct type.
    return Region.decode(encoded)


# Register all the region classes with the same constructor and representer
if yaml:
    for region_class in (ConvexPolygon, Ellipse, Circle, Box, UnionRegion, IntersectionRegion):
        yaml.add_representer(region_class, region_representer)

        for loader in YamlLoaders:
            yaml.add_constructor(f"lsst.sphgeom.{region_class.__name__}", region_constructor, Loader=loader)


# Pixelization schemes


def pixel_representer(dumper, data):
    """Represent a pixelization in YAML.

    Stored as the pixelization level in a mapping with a single key
    ``level``.
    """
    return dumper.represent_mapping(f"lsst.sphgeom.{type(data).__name__}", {"level": data.getLevel()})


def pixel_constructor(loader, node):
    """Construct a pixelization object from YAML."""
    mapping = loader.construct_mapping(node)

    className = node.tag
    pixelMap = {
        "lsst.sphgeom.Q3cPixelization": Q3cPixelization,
        "lsst.sphgeom.Mq3cPixelization": Mq3cPixelization,
        "lsst.sphgeom.HtmPixelization": HtmPixelization,
        "lsst.sphgeom.HealpixPixelization": HealpixPixelization,
    }

    if className not in pixelMap:
        raise RuntimeError(
            f"Encountered unexpected class {className} associated with sphgeom pixelization YAML constructor"
        )

    return pixelMap[className](mapping["level"])


# All the pixelization schemes use the same approach with getLevel
if yaml:
    for pixelSchemeCls in (HtmPixelization, Q3cPixelization, Mq3cPixelization, HealpixPixelization):
        yaml.add_representer(pixelSchemeCls, pixel_representer)
        for loader in YamlLoaders:
            yaml.add_constructor(f"lsst.sphgeom.{pixelSchemeCls.__name__}", pixel_constructor, Loader=loader)



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import Matrix3d, Vector3d


class Matrix3dTestCase(unittest.TestCase):
    """Test 3D Matrix."""

    def testConstruction(self):
        self.assertEqual(Matrix3d().getSquaredNorm(), 0)
        self.assertEqual(Matrix3d(1, 0, 0, 0, 1, 0, 0, 0, 1), Matrix3d(1))

    def testComparison(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        self.assertEqual(m, m)
        self.assertNotEqual(m, Matrix3d(2))

    def testIter(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        r = 0
        for row in m:
            self.assertEqual(row, Vector3d(*range(r * 3 + 1, r * 3 + 4)))
            r += 1

    def testAccess(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        self.assertEqual(len(m), 3)
        self.assertEqual(m[0], Vector3d(1, 2, 3))
        self.assertEqual(m[1], Vector3d(4, 5, 6))
        self.assertEqual(m[2], Vector3d(7, 8, 9))
        self.assertEqual(m[0], m.getRow(0))
        self.assertEqual(m[1], m.getRow(1))
        self.assertEqual(m[2], m.getRow(2))
        n = m.transpose()
        self.assertEqual(m.getColumn(0), n[0])
        self.assertEqual(m.getColumn(1), n[1])
        self.assertEqual(m.getColumn(2), n[2])
        for r in range(3):
            for c in range(3):
                self.assertEqual(m[r, c], r * 3 + c + 1)
        for i in (-4, 3):
            with self.assertRaises(IndexError):
                m.getRow(i)
            with self.assertRaises(IndexError):
                m.getColumn(i)
        for i in [-4, 3, (-4, 1), (3, 1), (1, -4), (1, 3)]:
            with self.assertRaises(IndexError):
                m[i]

    def testInner(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        i = Matrix3d(1)
        self.assertEqual(m.inner(i), i.inner(m))
        self.assertEqual(m.inner(i), 15.0)

    def testNorm(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 0, 2)
        self.assertEqual(m.getSquaredNorm(), 144.0)
        self.assertEqual(m.getNorm(), 12.0)

    def testArithmetic(self):
        v = Vector3d(1, 2, 3)
        m = Matrix3d(1, -1, 0, 1, 1, 0, 0, 0, 1)
        n = Matrix3d(1, 1, 0, -1, 1, 0, 0, 0, 1)
        self.assertEqual(n * (m * v), Vector3d(2, 4, 3))
        self.assertEqual(m + m, m * Matrix3d(2))
        self.assertEqual(m, m * Matrix3d(2) - m)

    def testCwiseProduct(self):
        m = Matrix3d(1, 2, 3, 4, 1, 6, 7, 8, 1)
        self.assertEqual(m.cwiseProduct(Matrix3d(2)), Matrix3d(2))

    def testTranspose(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        n = Matrix3d(1, 4, 7, 2, 5, 8, 3, 6, 9)
        self.assertEqual(m.transpose(), n)
        self.assertEqual(m.transpose().transpose(), m)

    def testInverse(self):
        m = Matrix3d(4, 4, 4, -1, 1, 0, 1, -1, -1)
        n = Matrix3d(0.125, 0, 0.5, 0.125, 1, 0.5, 0, -1, -1)
        i = m.inverse()
        self.assertEqual(i, n)
        self.assertEqual(i.inverse(), m)
        self.assertEqual(i * m, Matrix3d(1))
        self.assertEqual(m * i, Matrix3d(1))

    def testString(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        self.assertEqual(str(m), "[[1.0, 2.0, 3.0],\n [4.0, 5.0, 6.0],\n [7.0, 8.0, 9.0]]")
        self.assertEqual(repr(m), "Matrix3d(1.0, 2.0, 3.0,\n         4.0, 5.0, 6.0,\n         7.0, 8.0, 9.0)")
        self.assertEqual(m, eval(repr(m), {"Matrix3d": Matrix3d}))

    def testPickle(self):
        m = Matrix3d(1, 2, 3, 4, 5, 6, 7, 8, 9)
        n = pickle.loads(pickle.dumps(m))
        self.assertEqual(m, n)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import CONTAINS, DISJOINT, Interval1d


class Interval1dTestCase(unittest.TestCase):
    """Test 1D intervals."""

    def testConstruction(self):
        i = Interval1d(1)
        self.assertEqual(i.getA(), i.getB())
        self.assertEqual(i.getA(), 1)
        i = Interval1d(1, 2)
        self.assertEqual(i, Interval1d(1, 2))
        self.assertTrue(Interval1d.empty().isEmpty())

    def testComparisonOperators(self):
        self.assertEqual(Interval1d(1), Interval1d(1, 1))
        self.assertEqual(Interval1d(1), 1)
        self.assertNotEqual(Interval1d(1, 1), Interval1d(2, 2))
        self.assertNotEqual(Interval1d(2, 2), 1)

    def testCenterAndSize(self):
        i = Interval1d(1, 2)
        self.assertEqual(i.getSize(), 1)
        self.assertEqual(i.getCenter(), 1.5)

    def testRelationships(self):
        i02 = Interval1d(0, 2)
        i13 = Interval1d(1, 3)
        i46 = Interval1d(4, 6)
        i06 = Interval1d(0, 6)
        self.assertTrue(i02.contains(1))
        self.assertTrue(i02.contains(Interval1d(0.5, 1.5)))
        self.assertTrue(i02.isDisjointFrom(3))
        self.assertTrue(i02.isDisjointFrom(i46))
        self.assertTrue(i02.intersects(1))
        self.assertTrue(i02.intersects(i13))
        self.assertTrue(Interval1d(1, 1).isWithin(i02))
        self.assertTrue(i02.isWithin(i06))
        r = i02.relate(1)
        self.assertEqual(r, CONTAINS)
        r = i46.relate(i02)
        self.assertEqual(r, DISJOINT)

    def testExpandingAndClipping(self):
        a = Interval1d(1, 2)
        b = a.expandedTo(3).expandedTo(Interval1d(2, 4)).clippedTo(Interval1d(0, 2)).clippedTo(1)
        a.expandTo(3).expandTo(Interval1d(2, 4))
        a.clipTo(Interval1d(0, 2)).clipTo(1)
        self.assertEqual(a, b)
        self.assertEqual(a, 1)

    def testDilationAndErosion(self):
        a = Interval1d(1, 3)
        b = a.dilatedBy(1).erodedBy(2)
        a.dilateBy(1).erodeBy(2)
        self.assertEqual(a, b)
        self.assertEqual(a, 2)

    def testString(self):
        i = Interval1d(1, 2)
        self.assertEqual(str(i), "[1.0, 2.0]")
        self.assertEqual(repr(i), "Interval1d(1.0, 2.0)")
        self.assertEqual(i, eval(repr(i), {"Interval1d": Interval1d}))

    def testPickle(self):
        a = Interval1d(1.5, 3.5)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)


if __name__ == "__main__":
    unittest.main()



"""
Basic setuptools description.

This is not a complete definition.

* Version number is not correct.
* The shared library and include files are not installed.  This makes it
  unusable with other python packages that directly reference the C++
  interface.
"""

import glob

from pybind11.setup_helpers import ParallelCompile, Pybind11Extension, build_ext
from setuptools import setup

# Optional multithreaded build.
ParallelCompile("NPY_NUM_BUILD_JOBS").install()

# Find the source code -- we can combine it into a single module
pybind_src = sorted(glob.glob("python/lsst/sphgeom/*.cc"))
cpp_src = sorted(glob.glob("src/*.cc"))

# Very inefficient approach since this compiles the maing sphgeom
# library code for every extension rather than building everything once
ext_modules = [
    Pybind11Extension("lsst.sphgeom._sphgeom", sorted(cpp_src + pybind_src), include_dirs=["include"])
]

setup(
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import pickle
import unittest

from lsst.sphgeom import Angle, LonLat, UnitVector3d, Vector3d


class UnitVector3dTestCase(unittest.TestCase):
    """Test 3D UnitVector."""

    def testConstruction(self):
        v = Vector3d(1, 1, 1)
        u = UnitVector3d.orthogonalTo(v)
        self.assertAlmostEqual(u.dot(v), 0.0, places=15)
        a = UnitVector3d(1, 1, 1)
        self.assertEqual(a, UnitVector3d(Vector3d(1, 1, 1)))
        self.assertAlmostEqual(a.x(), math.sqrt(3.0) / 3.0, places=15)
        self.assertAlmostEqual(a.y(), math.sqrt(3.0) / 3.0, places=15)
        self.assertAlmostEqual(a.z(), math.sqrt(3.0) / 3.0, places=15)
        b = UnitVector3d(Angle.fromDegrees(45), Angle.fromDegrees(45))
        self.assertEqual(b, UnitVector3d(LonLat.fromDegrees(45, 45)))
        self.assertAlmostEqual(b.x(), 0.5, places=15)
        self.assertAlmostEqual(b.y(), 0.5, places=15)
        self.assertAlmostEqual(b.z(), 0.5 * math.sqrt(2.0), places=15)
        c = UnitVector3d.northFrom(b)
        d = UnitVector3d(LonLat.fromDegrees(225, 45))
        self.assertAlmostEqual(c.x(), d.x(), places=15)
        self.assertAlmostEqual(c.y(), d.y(), places=15)
        self.assertAlmostEqual(c.z(), d.z(), places=15)

    def testComparison(self):
        self.assertEqual(UnitVector3d.X(), UnitVector3d.X())
        self.assertNotEqual(UnitVector3d.Y(), UnitVector3d.Z())

    def testAccess(self):
        v = UnitVector3d.X()
        self.assertEqual(len(v), 3)
        self.assertEqual(v[0], 1)
        self.assertEqual(v[1], 0)
        self.assertEqual(v[2], 0)
        self.assertEqual(v[-3], 1)
        self.assertEqual(v[-2], 0)
        self.assertEqual(v[-1], 0)
        with self.assertRaises(IndexError):
            v[-4]
        with self.assertRaises(IndexError):
            v[3]

    def testDot(self):
        self.assertEqual(UnitVector3d.X().dot(UnitVector3d.Z()), 0)

    def testCross(self):
        self.assertEqual(UnitVector3d.X().cross(UnitVector3d.Y()), Vector3d(0, 0, 1))
        self.assertEqual(UnitVector3d.X().robustCross(UnitVector3d.Y()), Vector3d(0, 0, 2))

    def testArithmeticOperators(self):
        self.assertEqual(-UnitVector3d.X(), UnitVector3d(-1, 0, 0))
        self.assertEqual(UnitVector3d.X() - UnitVector3d.X(), Vector3d(0, 0, 0))
        self.assertEqual(UnitVector3d.X() + UnitVector3d(1, 0, 0), UnitVector3d.X() * 2)
        self.assertEqual(UnitVector3d.Y() - Vector3d(0, 0.5, 0), UnitVector3d.Y() / 2)
        self.assertEqual(UnitVector3d.Z().cwiseProduct(Vector3d(2, 3, 4)), Vector3d(0, 0, 4))

    def testRotation(self):
        v = UnitVector3d.Y().rotatedAround(UnitVector3d.X(), Angle(0.5 * math.pi))
        self.assertAlmostEqual(v.x(), 0.0, places=15)
        self.assertAlmostEqual(v.y(), 0.0, places=15)
        self.assertAlmostEqual(v.z(), 1.0, places=15)

    def testString(self):
        v = UnitVector3d.X()
        self.assertEqual(str(v), "[1.0, 0.0, 0.0]")
        self.assertEqual(repr(v), "UnitVector3d(1.0, 0.0, 0.0)")
        self.assertEqual(v, eval(repr(v), {"UnitVector3d": UnitVector3d}))

    def testPickle(self):
        v = UnitVector3d(1, 1, 1)
        w = pickle.loads(pickle.dumps(v, pickle.HIGHEST_PROTOCOL))
        self.assertEqual(v, w)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import math
import pickle
import unittest

from lsst.sphgeom import Angle, UnitVector3d, Vector3d


class Vector3dTestCase(unittest.TestCase):
    """Test 3D vector."""

    def testConstruction(self):
        v = Vector3d(1, 2, 3)
        self.assertEqual(v.x(), 1)
        self.assertEqual(v.y(), 2)
        self.assertEqual(v.z(), 3)

    def testComparison(self):
        v = Vector3d(1, 2, 3)
        self.assertEqual(v, Vector3d(1, 2, 3))
        self.assertNotEqual(v, Vector3d(1, 2, 4))

    def testIsZero(self):
        self.assertTrue(Vector3d(0, 0, 0).isZero())
        self.assertFalse(Vector3d(0, 0, 1).isZero())

    def testAccess(self):
        v = Vector3d(1, 2, 3)
        self.assertEqual(len(v), 3)
        self.assertEqual(v[0], 1)
        self.assertEqual(v[1], 2)
        self.assertEqual(v[2], 3)
        self.assertEqual(v[-3], 1)
        self.assertEqual(v[-2], 2)
        self.assertEqual(v[-1], 3)
        with self.assertRaises(IndexError):
            v[-4]
        with self.assertRaises(IndexError):
            v[3]

    def testNormal(self):
        v = Vector3d(0, 2, 0)
        self.assertEqual(v.getSquaredNorm(), 4)
        self.assertEqual(v.getNorm(), 2)
        v.normalize()
        self.assertTrue(v.isNormalized())
        self.assertEqual(v, Vector3d(0, 1, 0))

    def testDot(self):
        x = Vector3d(1, 0, 0)
        y = Vector3d(0, 1, 0)
        self.assertEqual(x.dot(y), 0)

    def testCross(self):
        x = Vector3d(1, 0, 0)
        y = Vector3d(0, 1, 0)
        self.assertEqual(x.cross(y), Vector3d(0, 0, 1))

    def testArithmeticOperators(self):
        self.assertEqual(-Vector3d(1, 1, 1), Vector3d(-1, -1, -1))
        self.assertEqual(Vector3d(1, 1, 1) * 2, Vector3d(2, 2, 2))
        self.assertEqual(Vector3d(2, 2, 2) / 2, Vector3d(1, 1, 1))
        self.assertEqual(Vector3d(1, 1, 1) + Vector3d(1, 1, 1), Vector3d(2, 2, 2))
        self.assertEqual(Vector3d(1, 1, 1) - Vector3d(1, 1, 1), Vector3d())
        v = Vector3d(1, 1, 1)
        v += Vector3d(3, 3, 3)
        v -= Vector3d(2, 2, 2)
        v *= 2.0
        v /= 4.0
        self.assertEqual(v, Vector3d(1, 1, 1))
        self.assertEqual(v.cwiseProduct(Vector3d(2, 3, 4)), Vector3d(2, 3, 4))

    def testRotation(self):
        v = Vector3d(0, 1, 0).rotatedAround(UnitVector3d.X(), Angle(0.5 * math.pi))
        self.assertAlmostEqual(v.x(), 0.0, places=15)
        self.assertAlmostEqual(v.y(), 0.0, places=15)
        self.assertAlmostEqual(v.z(), 1.0, places=15)

    def testString(self):
        v = Vector3d(1, 0, 0)
        self.assertEqual(str(v), "[1.0, 0.0, 0.0]")
        self.assertEqual(repr(v), "Vector3d(1.0, 0.0, 0.0)")
        self.assertEqual(v, eval(repr(v), {"Vector3d": Vector3d}))

    def testPickle(self):
        v = Vector3d(1, 2, 3)
        w = pickle.loads(pickle.dumps(v))
        self.assertEqual(v, w)


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pickle
import unittest

from lsst.sphgeom import CONTAINS, DISJOINT, Angle, AngleInterval, NormalizedAngle, NormalizedAngleInterval


class IntervalTests:
    """Test intervals."""

    def testConstruction(self):
        i = self.Interval(self.Scalar(1))
        self.assertEqual(i.getA(), i.getB())
        self.assertEqual(i.getA(), self.Scalar(1))
        i = self.Interval(self.Scalar(1), self.Scalar(2))
        self.assertEqual(i, self.Interval.fromRadians(1, 2))
        self.assertTrue(self.Interval.empty().isEmpty())

    def testComparisonOperators(self):
        self.assertEqual(self.Interval(self.Scalar(1)), self.Interval.fromRadians(1, 1))
        self.assertEqual(self.Interval(self.Scalar(1)), self.Scalar(1))
        self.assertNotEqual(self.Interval.fromDegrees(1, 1), self.Interval.fromRadians(1, 1))
        self.assertNotEqual(self.Interval.fromDegrees(2, 2), self.Scalar(1))

    def testCenterAndSize(self):
        a = self.Interval.fromRadians(1, 2)
        self.assertEqual(a.getSize(), self.Scalar(1))
        self.assertEqual(a.getCenter(), self.Scalar(1.5))

    def testRelationships(self):
        a02 = self.Interval.fromRadians(0, 2)
        a13 = self.Interval.fromRadians(1, 3)
        a46 = self.Interval.fromRadians(4, 6)
        a06 = self.Interval.fromRadians(0, 6)
        self.assertTrue(a02.contains(self.Scalar(1)))
        self.assertTrue(a02.contains(self.Interval.fromRadians(0.5, 1.5)))
        self.assertTrue(a02.isDisjointFrom(self.Scalar(3)))
        self.assertTrue(a02.isDisjointFrom(a46))
        self.assertTrue(a02.intersects(self.Scalar(1)))
        self.assertTrue(a02.intersects(a13))
        self.assertTrue(self.Interval.fromRadians(1, 1).isWithin(a02))
        self.assertTrue(a02.isWithin(a06))
        r = a02.relate(self.Scalar(1))
        self.assertEqual(r, CONTAINS)
        r = a46.relate(a02)
        self.assertEqual(r, DISJOINT)

    def testExpandingAndClipping(self):
        a = self.Interval.fromRadians(1, 2)
        b = (
            a.expandedTo(self.Scalar(3))
            .expandedTo(self.Interval.fromRadians(2, 4))
            .clippedTo(self.Interval.fromRadians(0, 2))
            .clippedTo(self.Scalar(1))
        )
        a.expandTo(self.Scalar(3)).expandTo(self.Interval.fromRadians(2, 4))
        a.clipTo(self.Interval.fromRadians(0, 2)).clipTo(self.Scalar(1))
        self.assertEqual(a, b)
        self.assertEqual(a, self.Scalar(1))

    def testDilationAndErosion(self):
        a = self.Interval.fromRadians(1, 3)
        b = a.dilatedBy(self.Scalar(1)).erodedBy(self.Scalar(2))
        a.dilateBy(self.Scalar(1)).erodeBy(self.Scalar(2))
        self.assertEqual(a, b)
        self.assertEqual(a, self.Scalar(2))

    def testString(self):
        a = self.Interval.fromRadians(0.5, 1.5)
        self.assertEqual(str(a), "[0.5, 1.5]")
        self.assertEqual(repr(a), self.Interval.__name__ + ".fromRadians(0.5, 1.5)")
        self.assertEqual(a, eval(repr(a), {self.Interval.__name__: self.Interval}))

    def testPickle(self):
        a = self.Interval.fromRadians(1, 3)
        b = pickle.loads(pickle.dumps(a))
        self.assertEqual(a, b)


class AngleIntervalTestCase(unittest.TestCase, IntervalTests):
    """Angle interval test cases."""

    def setUp(self):
        self.Interval = AngleInterval
        self.Scalar = Angle


class NormalizedAngleIntervalTestCase(unittest.TestCase, IntervalTests):
    """Normalized angle interval test cases."""

    def setUp(self):
        self.Interval = NormalizedAngleInterval
        self.Scalar = NormalizedAngle


if __name__ == "__main__":
    unittest.main()



# This file is part of sphgeom.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Extend any of the C++ Python classes by adding additional methods."""

# Nothing to export.
__all__ = []

import math
import sys
import typing

from ._sphgeom import Angle, Box, Circle, ConvexPolygon, LonLat, Region, UnitVector3d

# Copy and paste from lsst.utils.wrappers:
# * INTRINSIC_SPECIAL_ATTRIBUTES
# * isAttributeSafeToTransfer
# * continueClass
_INTRINSIC_SPECIAL_ATTRIBUTES = frozenset(
    (
        "__qualname__",
        "__module__",
        "__metaclass__",
        "__dict__",
        "__weakref__",
        "__class__",
        "__subclasshook__",
        "__name__",
        "__doc__",
    )
)


def _isAttributeSafeToTransfer(name: str, value: typing.Any) -> bool:
    if name.startswith("__") and (
        value is getattr(object, name, None) or name in _INTRINSIC_SPECIAL_ATTRIBUTES
    ):
        return False
    return True


def _continueClass(cls):
    orig = getattr(sys.modules[cls.__module__], cls.__name__)
    for name in dir(cls):
        # Common descriptors like classmethod and staticmethod can only be
        # accessed without invoking their magic if we use __dict__; if we use
        # getattr on those we'll get e.g. a bound method instance on the dummy
        # class rather than a classmethod instance we can put on the target
        # class.
        attr = cls.__dict__.get(name, None) or getattr(cls, name)
        if _isAttributeSafeToTransfer(name, attr):
            setattr(orig, name, attr)
    return orig


def _inf_to_limit(value: float, min: float, max: float) -> float:
    """Map a value to a fixed range if infinite."""
    if not math.isinf(value):
        return value
    if value > 0.0:
        return max
    return min


def _inf_to_lat(lat: float) -> float:
    """Map latitude +Inf to +90 and -Inf to -90 degrees."""
    return _inf_to_limit(lat, -90.0, 90.0)


def _inf_to_lon(lat: float) -> float:
    """Map longitude +Inf to +360 and -Inf to 0 degrees."""
    return _inf_to_limit(lat, 0.0, 360.0)


@_continueClass
class Region:
    """A minimal interface for 2-dimensional regions on the unit sphere."""

    @classmethod
    def from_ivoa_pos(cls, pos: str) -> Region:
        """Create a Region from an IVOA POS string.

        Parameters
        ----------
        pos : `str`
            A string using the IVOA SIAv2 POS syntax.

        Returns
        -------
        region : `Region`
            A region equivalent to the POS string.

        Notes
        -----
        See
        https://ivoa.net/documents/SIA/20151223/REC-SIA-2.0-20151223.html#toc12
        for a description of the POS parameter but in summary the options are:

        * ``CIRCLE <longitude> <latitude> <radius>``
        * ``RANGE <longitude1> <longitude2> <latitude1> <latitude2>``
        * ``POLYGON <longitude1> <latitude1> ... (at least 3 pairs)``

        Units are degrees in all coordinates.
        """
        shape, *coordinates = pos.split()
        coordinates = tuple(float(c) for c in coordinates)
        n_floats = len(coordinates)
        if shape == "CIRCLE":
            if n_floats != 3:
                raise ValueError(f"CIRCLE requires 3 numbers but got {n_floats} in '{pos}'.")
            center = LonLat.fromDegrees(coordinates[0], coordinates[1])
            radius = Angle.fromDegrees(coordinates[2])
            return Circle(UnitVector3d(center), radius)

        if shape == "RANGE":
            if n_floats != 4:
                raise ValueError(f"RANGE requires 4 numbers but got {n_floats} in '{pos}'.")
            # POS allows +Inf and -Inf in ranges. These are not allowed by
            # Box and so must be converted.
            return Box(
                LonLat.fromDegrees(_inf_to_lon(coordinates[0]), _inf_to_lat(coordinates[2])),
                LonLat.fromDegrees(_inf_to_lon(coordinates[1]), _inf_to_lat(coordinates[3])),
            )

        if shape == "POLYGON":
            if n_floats % 2 != 0:
                raise ValueError(f"POLYGON requires even number of floats but got {n_floats} in '{pos}'.")
            if n_floats < 6:
                raise ValueError(
                    f"POLYGON specification requires at least 3 coordinates, got {n_floats // 2} in '{pos}'"
                )
            # Coordinates are x1, y1, x2, y2, x3, y3...
            # Get pairs by skipping every other value.
            pairs = list(zip(coordinates[0::2], coordinates[1::2], strict=True))
            vertices = [LonLat.fromDegrees(lon, lat) for lon, lat in pairs]
            return ConvexPolygon([UnitVector3d(c) for c in vertices])

        raise ValueError(f"Unrecognized shape in POS string '{pos}'")

    def to_ivoa_pos(self) -> str:
        """Represent the region as an IVOA POS string.

        Returns
        -------
        pos : `str`
            The region in ``POS`` format.
        """
        raise NotImplementedError("This region can not be converted to an IVOA POS string.")


@_continueClass
class Circle:  # noqa: F811
    """A circular region on the unit sphere that contains its boundary."""

    def to_ivoa_pos(self) -> str:
        # Docstring inherited.
        center = LonLat(self.getCenter())
        lon = center.getLon().asDegrees()
        lat = center.getLat().asDegrees()
        rad = self.getOpeningAngle().asDegrees()
        return f"CIRCLE {lon} {lat} {rad}"


@_continueClass
class Box:  # noqa: F811
    """A rectangle in spherical coordinate space that contains its boundary."""

    def to_ivoa_pos(self) -> str:
        # Docstring inherited.
        lon_range = self.getLon()
        lat_range = self.getLat()

        lon1 = lon_range.getA().asDegrees()
        lon2 = lon_range.getB().asDegrees()
        lat1 = lat_range.getA().asDegrees()
        lat2 = lat_range.getB().asDegrees()

        # Do not attempt to map to +/- Inf -- there is no way to know if
        # that is any better than 0. -> 360.
        return f"RANGE {lon1} {lon2} {lat1} {lat2}"


@_continueClass
class ConvexPolygon:  # noqa: F811
    """A rectangle in spherical coordinate space that contains its boundary."""

    def to_ivoa_pos(self) -> str:
        # Docstring inherited.
        coords = (LonLat(v) for v in self.getVertices())
        coord_strings = [f"{c.getLon().asDegrees()} {c.getLat().asDegrees()}" for c in coords]

        return f"POLYGON {' '.join(coord_strings)}"
